# the inclusion of the tests module is not meant to offer best practices for
# testing in general, but rather to support the `find_packages` example in
# setup.py that excludes installing the "tests" package


from tko.util.text  import Text, Token

class TestSimple:
    def test_token_creation(self):
        token: Token = Token("text1","g")
        assert token.text == "text1"
        assert token.fmt == "g"

    def test_getitem(self):
        sentence = Text() + Token("tex", "g") + Token("gu", "r")
        assert sentence[0] == Token("t", "g")
        assert sentence[1] == Token("e", "g")
        assert str(sentence[2]) == str(Token("x", "g"))
        assert sentence[2] == Token("x", "g")
        assert sentence[3] == Token("g", "r")
        assert sentence[4] == Token("u", "r")


    def test_token_addition(self):
        token1 = Token("text1", "g")
        token2 = Token("text2", "r")
        sentence = token1 + token2
        assert len(sentence) == 10
        assert sentence[0] == Token("t", "g")
        assert "".join([x.fmt for x in sentence.get_data()]) == "gggggrrrrr"
        assert "".join([x.text for x in sentence.get_data()]) == "text1text2"

    def test_token_equality(self):
        token1 = Token("text1", "fmt1")
        token2 = Token("text1", "fmt1")
        token3 = Token("text2", "fmt2")
        assert token1 == token2
        assert token1 != token3

    def test_token_length(self):
        token = Token("text1", "fmt1")
        assert len(token) == 5

    def test_sentence_creation(self):
        sentence = Text()
        assert len(sentence) == 0

    def test_sentence_addition(self):
        sentence1 = Text() + Token("text1", "")
        sentence2 = Text() + Token("text2", "r")
        sentence3 = sentence1 + sentence2
        assert len(sentence3) == 10
        assert "".join([x.fmt for x in sentence3.get_data()]) == "rrrrr"
        assert "".join([x.text for x in sentence3.get_data()]) == "text1text2"
        assert sentence3.resume() == [Token("text1"), Token("text2", "r")]
        sentence4 = sentence1 + "text3"
        assert len(sentence4) == 10
        assert sentence4.resume() == [Token("text1text3")]

    def test_sentence_equality(self):
        sentence1 = Text() + Token("text1", "fmt1") + Token("text2", "fmt2")
        sentence2 = Text() + Token("text1", "fmt1") + Token("text2", "fmt2")
        sentence3 = Text() + Token("text3", "fmt3")
        assert sentence1 == sentence2
        assert sentence1 != sentence3

    def test_sentence_len(self):
        sentence = Text() + Token("text1", "fmt1") + Token("text2", "fmt2")
        assert sentence.len() == 10

    def test_sentence_trim_end(self):
        sentence = Text() + Token("text1", "r") + Token("text2", "g")
        trimmed_sentence = sentence.trim_end(7)
        assert trimmed_sentence.len() == 7
        assert trimmed_sentence.resume() == [Token("text1", "r"), Token("te", "g")]

    def test_text_replace1(self):
        bar = Text() + Token("text1foo", "g") + Token("text2", "b")
        bar.replace("t", Token("x", "r"))
        assert "".join([x.text for x in bar.get_data()]) == "xexx1fooxexx2"
        assert "".join([x.fmt for x in bar.get_data()]) == "rggrggggrbbrb"

    def test_text_replace2(self):
        bar = Text.format("123456789")
        bar.replace("123", Token("abc", "r"))
        assert str(bar) == str(Text.format("{r}456789", "abc"))

        bar = Text.format("123456789")
        bar.replace("123", Token("ab", "r"))
        assert str(bar) == str(Text.format("{r}456789", "ab"))

        bar = Text.format("0123412789")
        bar.replace("12", Token("a", "r"))
        assert str(bar) == str(Text.format("0{r}34{r}789", "a", "a"))

        bar = Text.format("0123412789")
        bar.replace("12", Token("abcd", "r"))
        assert str(bar) == str(Text.format("0{r}34{r}789", "abcd", "abcd"))


    def test_resume1(self):
        sentence = Text() + Token("text1", "g") + Token("text2", "g")
        assert sentence.resume() == [Token("text1text2", "g")]

    def test_resume2(self):
        sentence = Text() + Token("text1", "g") + Token("text2", "r")
        assert sentence.resume() == [Token("text1", "g"), Token("text2", "r")]
    

    def test_split_1(self):
        sentence = Text() + Token("text1", "g") + Token("ext2", "r")
        assert sentence.split("t") == [Text(), Text().addf("g", "ex"), Text().addf("g", "1").addf("r", "ex"), Text().addf("r", "2")]

    def test_slice(self):
        sentence = Text() + Token("text1", "g") + Token("text2", "r")
        assert sentence.slice(1, -1) == Text().addf("g", "ext1") + Token("text", "r")
        print(sentence.slice(1, -1))
        assert sentence.slice(1) == Text().addf("g", "ext1") + Token("text2", "r")
        assert sentence.slice() == Text().addf("g", "text1") + Token("text2", "r")

    def test_format(self):
        text1 = Text.format("brasil é {g} e {y}", "verde", "amarelo")
        text2 = Text.format("brasil é ").addf("g", "verde").add(" e ").addf("y", "amarelo")
        assert str(text1) == str(text2)

        text1 = Text.format("ceu é {b}, ouro é {}", "azul", Token("amarelo", "y"))
        text2 = Text().add("ceu é ").addf("b", "azul").add(", ouro é ").addf("y", "amarelo")
        assert str(text1) == str(text2)

        text1 = Text.format("sangue é {}, mar é {c}", Text.format("{r}", "vermelho"), Token("ciano", "r"))
        text2 = Text().add("sangue é ").addf("r", "vermelho").add(", mar é ").addf("c", "ciano")
        assert str(text1) == str(text2)

    def test_format_default(self):
        text1 = Text("b").add("banana").add(Token("uva", "g")).add(Text("r").add("ovo"))
        assert text1.resume() == [Token("banana", "b"), Token("uva", "g"), Token("ovo", "r")]

    def test_format_default2(self):
        text1 = Text("b") + "banana" + " " + "madura"
        text2 = Text("b").add("banana").add(" ").add("madura")
        assert str(text1) == str(text2)
        assert text1.resume() == [Token("banana madura", "b")]
        text1 = Text("b") + "banana" + Text("r") + Token("madura", "g")
        assert text1.resume() == [Token("banana", "b"), Token("madura", "g")]

    def test_format_1(self):
        text1 = Text.format("brasil é {g} e {y}", "verde", "amarelo", " e azul")
        text2 = Text().add("brasil é ").addf("g", "verde").add(" e ").addf("y", "amarelo").add(" e azul")
        assert str(text1) == str(text2)

        text3 = Text.format("o brasil é {g:verde} e {y:amarelo}.")
        text4 = Text.format("o brasil é {} e {}.", Token("verde", "g"), Token("amarelo", "y"))
        assert str(text3) == str(text4)

        text5 = Text() + "o brasil é " + Text("g") + "verde" + Text() + " e " + Text("y") + "amarelo"
        assert text5.resume() == [Token("o brasil é "), Token("verde", "g"), Token(" e "), Token("amarelo", "y")]

    def test_tuple(self):
        text1 = Text() + ("g", "brasil") + " é " + ("r", "lindo")
        print(text1)
        print("chiclete")



from tko.settings.settings import Settings
from tko.settings.repository import Repository
from tko.play.play import Play
from tko.settings.logger import Logger
from tko.util.text import Text

class CmdOpen:
    def __init__(self, settings: Settings):
        self.settings = settings
        self.need_update = False
        self.rep: Repository | None = None
        self.folder = ""

    def set_need_update(self):
        self.need_update = True

    def load_folder(self, folder: str):
        self.folder = folder
        self.rep = Repository(folder)
        if not self.rep.has_local_config_file():
            print(Text.format("{r}: O parâmetro para o comando {g} {y} deve a pasta onde você iniciou o repositório.", "Erro", "tko open", "<pasta>"))
            print(Text.format("{g}: Navegue ou passe o caminho até a pasta do repositório e tente novamente.", "Ação"))
            print(Text.format("{g}: Ou use {y} para criar um novo repositório.", "Ação","tko init --remote [fup|poo|ed]"))
            raise Warning(Text.format("{r}: {y} {}", "Erro:", folder, "não contém um repositório do tko"))
        self.rep.load_config().load_game()
        return self


    def execute(self):
        if self.rep is None:
            raise Warning("Repositório não encontrado")
        Logger.instance = Logger(self.rep)
        logger: Logger = Logger.get_instance()
        logger.set_log_files(self.rep.get_history_file(), self.rep.get_track_folder())
        logger.record_open()
        play = Play(self.settings, self.rep)
        if self.need_update:
            play.set_need_update()
        play.play()
        logger.record_quit()


    # @staticmethod
    # def choose_destiny(rep_alias):
    #     print(Text().add("Repositório local ainda não foi criado, onde você deseja criá-lo?"))
    #     here_cwd = os.path.join(os.getcwd(), rep_alias)
    #     qxcode = os.path.join(os.path.expanduser("~"), "qxcode", rep_alias)

    #     while True:
    #         print(Text().addf("r", "1").add(" - ").add(here_cwd))
    #         print(Text().addf("r", "2").add(" - ").add(qxcode))
    #         print(Text().addf("r", "3").add(" - ").add("Outra pasta"))
    #         print(Text().add("Default ").addf("r", "1").add(": "), end="")
    #         op = input()
    #         if op == "":
    #             op = "1"
    #         if op == "1":
    #             home_qxcode = here_cwd
    #             break
    #         if op == "2":
    #             home_qxcode = qxcode
    #             break
    #         if op == "3":
    #             print(Text().addf("y", "Navegue até o diretório desejado e execute o tko novamente."))
    #             exit(1)

    #     if not os.path.exists(home_qxcode):
    #         os.makedirs(home_qxcode)
    #     return home_qxcode


import yaml
import os
from tko.settings.log_action import LogAction
from tko.settings.log_info import LogInfo
import datetime
        
class TaskListener:
    TASK_CSV = "task_log.csv"
    def __init__(self):
        self.key_actions: dict[str, list[LogInfo]] = {}
        self.log_file: str | None = None
        self.format = '%Y-%m-%d %H:%M:%S'
        self.max_minutes = 60
        self.track_folder: str | None = None

    def set_track_folder(self, folder: str):
        self.track_folder = folder
        return self

    def listener(self, action: LogAction, new_entry: bool = False):
        decoder = LogInfo().decode(action)
        if decoder.key == "":
            return
        if decoder.key not in self.key_actions:
            self.key_actions[decoder.key] = [decoder]
        else:
            last_action = self.key_actions[decoder.key][-1]
            last_time = datetime.datetime.strptime(last_action.timestamp, self.format)
            current_time = datetime.datetime.strptime(decoder.timestamp, self.format)
            diff = current_time - last_time
            if diff.total_seconds() < self.max_minutes * 60:
                elapsed = last_action.elapsed + diff
            else:
                elapsed = last_action.elapsed
            decoder.elapsed = elapsed
            if decoder.type == LogAction.Type.SIZE.value:
                decoder.attempts = last_action.attempts + 1
            else:
                decoder.attempts = last_action.attempts
                decoder.lines = last_action.lines
            
            self.key_actions[decoder.key].append(decoder)

        if self.track_folder is None:
            return
        if not new_entry:
            return
        task_track_folder = os.path.join(self.track_folder, decoder.key)
        os.makedirs(task_track_folder, exist_ok=True)
        log_file = os.path.join(task_track_folder, self.TASK_CSV)
        if os.path.exists(log_file):
            with open(log_file, 'a') as file:
                file.write(str(decoder) + '\n')
        else:
            with open(log_file, 'w') as file:
                for entry in self.key_actions[decoder.key]:
                    file.write(str(entry) + '\n')

    def set_task_file(self, value: str):
        self.log_file = value
        return self
    
    def __str__(self) -> str: # yaml formatted output of key_actions
        output: dict[str, list[str]] = {}
        for key in sorted(self.key_actions.keys()):
            output[key] = []
            for action in self.key_actions[key]:
                output[key].append(str(action))
        return yaml.dump(output)

    def resume(self) -> dict[str, LogInfo]:
        output: dict[str, LogInfo] = {}
        for key in sorted(self.key_actions.keys()):
            entries = self.key_actions[key]
            info = LogInfo().set_key(key)
            info.set_elapsed(entries[-1].elapsed)
            info.set_attempts(len([x for x in entries if x.type == "TEST" or x.type == "FAIL"]))
            for action in entries:
                if action.coverage != -1:
                    info.set_coverage(action.coverage)
                if action.approach != -1:
                    info.set_autonomy(action.approach)
                if action.autonomy != -1:
                    info.set_skill(action.autonomy)
            output[key] = info
        return output
    
    def save_yaml(self):
        if self.log_file is None:
            return
        with open(self.log_file, 'w') as file:
            yaml.dump(self.key_actions, file)



import datetime
import os
import argparse
import csv
from tko.play.patch_history import PatchHistory
from tko.util.decoder import Decoder

class Track:
    def __init__(self):
        self.timestamp = ""
        self.files: list[str] = []
        self.result: str = ""

    def set_timestamp(self, timestamp: str):
        self.timestamp = timestamp
        return self
    
    def set_files(self, files: list[str]):
        self.files = [os.path.basename(f) for f in files]
        return self
    
    def set_result(self, result: str):
        self.result = result
        return self
    
    def track_to_column(self):
        return [self.timestamp, self.result, ";".join(self.files)]


class Tracker:

    def __init__(self):
        self._files: list[str] = []
        self._folder: str = ""
        self._result: str = "None"
        self.log_file = "track.csv"
        self.track_folder = ".track"
        self.extension = ".json"

    def set_files(self, files: list[str]):
        self._files = [os.path.abspath(f) for f in files]
        return self
    
    def set_result(self, result: str):
        self._result = result
        return self
    
    def set_percentage(self, percentage: int):
        self._result = "{}%".format(str(percentage).rjust(3, "0"))
        return self

    def get_log_full_path(self):
        return os.path.join(self._folder, self.log_file)

    def set_folder(self, folder: str):
        self._folder = folder
        return self
    
    # in format: YYYY-MM-DD HH:MM:SS
    def get_timestamp(self):
        return datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    # return timestamp of the last version of the file
    def save_file_with_timestamp_prefix(self, timestamp: str, file: str) -> tuple[str, bool, int]:
        filename = os.path.basename(file)
        ph = PatchHistory()
        json_file = os.path.join(self._folder, f"{filename}{self.extension}")
        ph.set_json_file(json_file)
        ph.load_json()

        content = Decoder.load(file)
        last_version = ph.store_version(timestamp, content)
        if last_version == timestamp:
            ph.save_json()
            return timestamp, True, len(content.splitlines())
        return last_version, False, len(content.splitlines())

    # return True if any file was changed
    def store(self) -> tuple[bool, int]:
        os.makedirs(self._folder, exist_ok=True)
        file_list = os.listdir(self._folder)
        file_list = [f for f in file_list if f.endswith(self.extension)]

        files_in_this_version: list[str] = []
        timestamp = self.get_timestamp()

        any_changes = False
        total_size = 0
        for file in self._files:
            stored, changed, size = self.save_file_with_timestamp_prefix(timestamp, file)
            total_size += size
            filename = os.path.basename(file)
            files_in_this_version.append(filename + ":" + stored)
            if changed:
                any_changes = True

        log_file = self.get_log_full_path()
        track = Track().set_timestamp(timestamp).set_files(files_in_this_version).set_result(self._result)
        with open(log_file, encoding="utf-8", mode="a") as f:
            writer = csv.writer(f)
            writer.writerow(track.track_to_column())
        return any_changes, total_size

    @staticmethod
    def main():
        parser = argparse.ArgumentParser(description="Track files changes.")
        parser.add_argument("files", metavar="files", type=str, nargs="+", help="files to be tracked.")
        args = parser.parse_args()
        
        tracker = Tracker().set_folder(".track").set_files(args.files)
        tracker.store()



from tko.play.floating import Floating

class FloatingManager:
    def __init__(self):
        self.input_layer: list[Floating] = []

    def add_input(self, floating: Floating):
        self.input_layer.append(floating)

    def draw_warnings(self):
        if len(self.input_layer) > 0 and self.input_layer[0].is_enable():
            self.input_layer[0].draw()

    def has_floating(self) -> bool:
        while len(self.input_layer) > 0 and not self.input_layer[0].is_enable():
            self.input_layer = self.input_layer[1:]
        return len(self.input_layer) > 0 and self.input_layer[0].is_enable()

    def get_input(self) -> int:
        return self.input_layer[0].get_input()



from tko.game.task_parser import TaskParser
from tko.game.cluster import Cluster
from tko.game.quest import Quest
from tko.game.task import Task
from tko.game.quest import Quest, QuestParser
from tko.game.task import Task
from tko.util.get_md_link import get_md_link
from tko.util.to_asc import uni_to_asc
import os

class GameBuilder:
    def __init__(self, filename: str, database_folder: str):
        self.filename = filename
        self.database_folder = database_folder
        self.ordered_clusters: list[str] = [] # ordered clusters
        self.clusters: dict[str, Cluster] = {}
        self.active_cluster: Cluster | None = None
        self.active_quest: Quest | None = None

    def build_from(self, content: str, language: str):
        self.__parse_file_content(content)
        self.__parse_database_for_user_tasks()
        self.__clear_empty_or_other_language(language)
        self.__create_requirements_pointers()
        self.__create_cross_references()
        return self


    def collect_tasks(self) -> dict[str, Task]:
        tasks: dict[str, Task] = {}
        for cluster in self.clusters.values():
            for quest in cluster.get_quests():
                for task in quest.get_tasks():
                    tasks[task.key] = task
        return tasks

    def collect_quests(self) -> dict[str, Quest]:
        quests: dict[str, Quest] = {}
        for cluster in self.clusters.values():
            for quest in cluster.get_quests():
                quests[quest.key] = quest
        return quests

    def __create_requirements_pointers(self):
        quests = self.collect_quests()
        # verificar se todas as quests requeridas existem e adicionar o ponteiro
        for q in quests.values():
            for r in q.requires:
                if r in quests:
                    q.requires_ptr.append(quests[r])
                else:
                    # print(f"keys: {self.quests.keys()}")
                    print(f"Quest\n{self.filename}:{q.line_number}\n{str(q)}\nrequer {r} que não existe")
                    exit(1)

    def __parse_database_for_user_tasks(self):
        folder = self.database_folder
        local_tasks_cluster: Cluster = Cluster(0, "Atividades Locais", "local_task_cluster")
        self.__add_cluster(local_tasks_cluster)
        tasks = self.collect_tasks()
        if not os.path.exists(folder):
            os.makedirs(folder)
        for entry in os.listdir(folder):
            path = os.path.join(folder, entry)
            if not os.path.isdir(path):
                continue
            if entry in tasks:
                continue
            # create task for folder
            task = Task()
            task.key = entry
            task.title = entry
            task.folder = path
            self.__add_task(task)

    def __parse_file_content(self, content: str):
        lines = content.splitlines()
        for line_num, line in enumerate(lines):
            cluster = self.__load_cluster(line, line_num)
            if cluster is not None:
                self.__add_cluster(cluster)
                continue
            
            quest = QuestParser().parse_quest(self.filename, line, line_num + 1)
            if quest is not None:
                self.__add_quest(quest)
                continue
            
            tp = TaskParser(self.filename, self.database_folder)
            task = tp.parse_line(line, line_num + 1)
            if task is not None:
                self.__add_task(task)

    def __get_active_cluster(self) -> Cluster:
        if self.active_cluster is None:
            raise Warning("Active Cluster Missing")
        return self.active_cluster

    def __get_active_quest(self) -> Quest:
        if self.active_quest is None:
            raise Warning("Active Quest Missing")
        return self.active_quest

    def __add_cluster(self, cluster: Cluster):
        self.clusters[cluster.key] = cluster
        self.ordered_clusters.append(cluster.key)
        self.active_cluster = cluster
        self.active_quest = None

    def __add_quest(self, quest: Quest):
        self.__get_active_cluster().add_quest(quest)
        self.active_quest = quest

    def __add_task(self, task: Task):
        if self.active_cluster is None:
            ckey = "sem_cluster"
            self.__add_cluster(Cluster(0, "Sem Cluster", ckey))
        if self.active_quest is None:
            qkey = self.__get_active_cluster().key + "_sem_quest"
            self.__add_quest(Quest("Sem Quest", qkey))
        self.__get_active_quest().add_task(task, self.filename)

    # se existir um cluster nessa linha, insere na lista de clusters e 
    # retorno o objeto cluster inserido
    def __load_cluster(self, line: str, line_num: int) -> None | Cluster:
        if not line.startswith("## "):
            return None
        line = line[3:]

        titulo = line
        tags_raw = ""
        if "<!--" in line:
            pieces = line.split("<!--")
            titulo = pieces[0]
            middle_end = pieces[1].split("-->")
            tags_raw = middle_end[0]
            titulo += middle_end[1]

        tags = [tag.strip() for tag in tags_raw.split(" ")]        
        key = uni_to_asc(get_md_link(titulo))
        try:
            color = [tag[2:] for tag in tags if tag.startswith("c:")][0]
        except IndexError as _e:
            color = None
        
        cluster = Cluster(line_num, titulo, key, color)

        if key in self.clusters.keys():
            c = self.clusters[key]
            print(f"Cluster {key} já existe")
            print(f"{self.filename}:{line_num}")
            print(f"{self.filename}:{c.line_number}")
            print("  " + str(c))
            print("  " + str(cluster))
            exit(1)
                
        self.clusters[key] = cluster
        self.ordered_clusters.append(key)
        return cluster

    def __clear_empty_or_other_language(self, language: str): #call before create_cross_references

        # apagando quests vazias da lista de quests
        for cluster in self.clusters.values():
            cluster.remove_empty_or_other_language(language)

        # apagando quests vazias dos clusters e clusters vazios
        ordered_clusters: list[str] = []
        clusters: dict[str, Cluster] = {}
        for key in self.ordered_clusters:
            cluster = self.clusters[key]
            if len(cluster.get_quests()) > 0:
                ordered_clusters.append(cluster.key)
                clusters[cluster.key] = cluster
        self.ordered_clusters = ordered_clusters
        self.clusters = clusters

    def __create_cross_references(self): #call after clear_empty
        for cluster in self.clusters.values():
            for quest in cluster.get_quests():
                quest.cluster_key = cluster.key
                for task in quest.get_tasks():
                    task.cluster_key = cluster.key
                    task.quest_key = quest.key


from tko.game.cluster import Cluster
from tko.game.quest import Quest
from tko.game.task import Task

class GameValidator:
    def __init__(self, filename: str, clusters: dict[str, Cluster], 
                 quests: dict[str, Quest], tasks: dict[str, Task]):
        self.filename = filename
        self.clusters: dict[str, Cluster] = clusters
        self.quests: dict[str, Quest] = {}
        self.tasks: dict[str, Task] = {}
        
    def validate(self):
        self.__validate_requirements()
        self.__check_cycle()
        return self

    def __validate_requirements(self):
        # verify is there are keys repeated between quests, tasks and groups
        keys = [c.key for c in self.clusters.values()] +\
               [k for k in self.quests.keys()] +\
               [k for k in self.tasks.keys()]

        # print chaves repetidas
        for k in keys:
            if keys.count(k) > 1:
                print(f"Chave repetida: {k}")
                exit(1)

        # trim titles
        for q in self.quests.values():
            q.title = q.title.strip()
        for c in self.clusters.values():
            c.title = c.title.strip()

        # verificar auto dependencia
        for q in self.quests.values():
            for r in q.requires:
                if q.key == r:
                    print(f"Erro: auto refência {q.line_number} {q.line}")
                    exit(1)


    # call after create_requirements_pointers
    def __check_cycle(self):
        def dfs(qx: Quest, visitedx: list[str]):
            if len(visitedx) > 0:
                if visitedx[0] == qx.key:
                    print(f"Cycle detected: {visitedx}")
                    exit(1)
            if q.key in visitedx:
                return
            visitedx.append(q.key)
            for r in q.requires_ptr:
                dfs(r, visitedx)

        for q in self.quests.values():
            visited: list[str] = []
            dfs(q, visited)


from tko.settings.repository import Repository
from tko.settings.settings import Settings
from tko.settings.logger import Logger
from tko.game.task import Task
from tko.play.week_graph import WeekGraph
import yaml # type: ignore
import os
import argparse
# from typing import override

class TaskResume:
    def __init__(self):
        self.coverage = 0
        self.autonomy = 0
        self.skill = 0
        self.elapsed = 0
        self.attempts = 0

    def set_coverage(self, value: int):
        self.coverage = value
        return self
    
    def set_autonomy(self, value: int):
        self.autonomy = value
        return self
    
    def set_skill(self, value: int):
        self.skill = value
        return self
    
    def set_elapsed(self, value: int):
        self.elapsed = value
        return self
    
    def set_attempts(self, value: int):
        self.attempts = value
        return self

    def to_dict(self):
        return {
            "coverage": self.coverage,
            "autonomy": self.autonomy,
            "skill": self.skill,
            "elapsed": self.elapsed,
            "attempts": self.attempts
        }

    # @override
    def __str__(self):
        return f"coverage:{self.coverage}, autonomy:{self.autonomy}, skill:{self.skill}, elapsed:{self.elapsed}, attempts:{self.attempts}"

class CmdRep:
    @staticmethod
    def check(args: argparse.Namespace):
        rep = Repository(args.folder).load_config().load_game()
        logger = Logger.get_instance()
        logger.set_log_files(rep.get_history_file(), rep.get_track_folder())

        output = logger.check_log_file_integrity()
        if len(output) == 0:
            print(f"Arquivo de log do repositório {rep} está íntegro.")
        else:
            print(f"Arquivo de log do repositório {rep} está corrompido.")
            print("Erros:")
            for error in output:
                print(f"- {error}")

    @staticmethod
    def resume(args: argparse.Namespace):
        rep = Repository(args.folder).load_config().load_game()
        logger = Logger.get_instance()
        logger.set_log_files(rep.get_history_file(), rep.get_track_folder())
        history_resume = logger.tasks.resume()
        repository_tasks: dict[str, Task] = rep.game.tasks

        tasks: dict[str, TaskResume] = {}
        for task in repository_tasks.values():
            if task.coverage != 0:
                tasks[task.key] = TaskResume().set_coverage(task.coverage).set_autonomy(task.approach).set_skill(task.autonomy)

        for key in history_resume:
            entry = history_resume[key]
            elapsed = entry.get_minutes()
            attempts = entry.attempts
            if elapsed > 0 or attempts > 0:
                if key not in tasks:
                    tasks[key] = TaskResume()
                tasks[key].set_elapsed(entry.get_minutes()).set_attempts(entry.attempts)
        
        tasks_str: dict[str, dict[str, int]] = {}
        for key in tasks:
            tasks_str[key] = tasks[key].to_dict()

        print (yaml.dump(tasks_str, sort_keys=False))

    @staticmethod
    def graph(args: argparse.Namespace):
        rep = Repository(args.folder).load_config().load_game()
        logger = Logger.get_instance()
        logger.set_log_files(rep.get_history_file(), rep.get_track_folder())
        wg = WeekGraph(100, 24, week_mode=True)
        image = wg.get_graph()
        week = wg.get_collected()
        pad = " " * 12
        print(f"{pad}{image[0]}")
        print(f"{pad}{image[1]}")
        image = image[2:]

        for i, line in enumerate(image):
            if i < len(week):
                index = i + 1
                index_pad = str(index).rjust(2, " ")
                print(f"{index_pad} - {week[i]}h ", end="")
            else:
                print(pad, end="")
            print(line)

        # for i in range(len(week)):
        #     print(f"{i} - {week[i]}h")

    @staticmethod
    def upgrade(args: argparse.Namespace):
        folder: str = args.folder
        if os.path.exists(os.path.join(folder, "rep.json")):
            os.rename(os.path.join(folder, "rep.json"), os.path.join(folder, "repository.json"))
        remote_folder = os.path.join(folder, "remote")
        os.makedirs(remote_folder, exist_ok=True)
        for entry in os.listdir(folder):
            path = os.path.join(folder, entry)
            if entry == "remote":
                continue
            if os.path.isdir(path):
                os.rename(path, os.path.join(remote_folder, entry))
        print(f"Repositório {folder} foi atualizado.")

    @staticmethod
    def list(_args: argparse.Namespace):
        settings = Settings()
        print(f"SettingsFile\n- {settings.settings_file}")
        print(str(settings))

    @staticmethod
    def add(args: argparse.Namespace):
        settings = Settings().set_alias_remote(args.alias, args.value)
        settings.save_settings()

    @staticmethod
    def rm(args: argparse.Namespace):
        sp = Settings()
        if args.alias in sp.dict_alias_remote:
            sp.dict_alias_remote.pop(args.alias)
            sp.save_settings()
        else:
            print("Repository not found.")

    @staticmethod
    def reset(args: argparse.Namespace):
        _ = args
        sp = Settings().reset()
        print(sp.settings_file)
        sp.save_settings()

    # @staticmethod
    # def graph(args):
    #     rep = Repository(args.folder).load_config().load_game()
    #     rep.game.check_cycle()
    #     Graph(rep.game).generate()



from __future__ import annotations

import argparse
import sys

from tko.cmds.cmd_rep import CmdRep
from tko.cmds.cmd_open import CmdOpen
from tko.cmds.cmd_down import CmdLineDown
from tko.cmds.cmd_run import Run
from tko.cmds.cmd_build import CmdBuild
from tko.cmds.cmd_config import CmdConfig, ConfigParams
from tko.enums.diff_mode import DiffMode
from tko.settings.logger import Logger
from tko.util.text import AnsiColor
from tko.enums.diff_count import DiffCount

from tko.util.param import Param
from tko.util.pattern import PatternLoader

from tko.settings.settings import Settings
from tko.cmds.cmd_diff import cmd_diff


from tko.settings.repository import Repository
from tko.util.raw_terminal import RawTerminal
from tko.util.symbols import symbols
from tko.settings.check_version import CheckVersion
from tko.settings.repository_starter import RepStarter

from tko.__init__ import __version__


class Main:
    @staticmethod
    def run(args: argparse.Namespace) -> int:
        if args.tui:
            return Main.tui(args)
        
        PatternLoader.pattern = args.pattern
        param = Param.Basic().set_index(args.index)
        if args.quiet:
            param.set_diff_count(DiffCount.QUIET)
        elif args.all:
            param.set_diff_count(DiffCount.ALL)
        else:
            param.set_diff_count(DiffCount.FIRST)

        if args.filter:
            param.set_filter(True)
        if args.compact:
            param.set_compact(True)

        settings = Settings()
        if not args.side and not args.down:
            param.set_diff_mode(settings.app.get_diff_mode())
        elif args.side:
            param.set_diff_mode(DiffMode.SIDE)
        elif args.down:
            param.set_diff_mode(DiffMode.DOWN)
        cmd_run = Run(settings, args.target_list, param)
        if args.eval:
            cmd_run.set_eval_mode()

        return cmd_run.execute()

    @staticmethod
    def tui(args: argparse.Namespace) -> int:
        PatternLoader.pattern = args.pattern
        param = Param.Basic().set_index(args.index)
        settings = Settings()
        param.set_diff_mode(settings.app.get_diff_mode())
        if args.filter:
            param.set_filter(True)
        cmd_run = Run(settings, args.target_list, param)
        cmd_run.set_curses()
        return cmd_run.execute()

    @staticmethod
    def build(args: argparse.Namespace):
        PatternLoader.pattern = args.pattern
        manip = Param.Manip().set_unlabel(args.unlabel).set_to_sort(args.sort).set_to_number(args.number)
        build = CmdBuild(args.target, args.target_list, manip)
        build.execute()

    @staticmethod
    def diff(args: argparse.Namespace):
        cmd_diff(args)

    @staticmethod
    def open(args: argparse.Namespace):
        settings = Settings()
        folder = args.folder
        rec_folder = Repository.rec_search_for_repo(folder)
        if rec_folder != "":
            folder = rec_folder
        action = CmdOpen(settings).load_folder(folder)
        if not CheckVersion().is_updated():
            action.set_need_update()
        action.execute()

    @staticmethod
    def down(args: argparse.Namespace):
        settings = Settings()
        folder = args.folder
        print("Folder: ", folder)
        rec_folder = Repository.rec_search_for_repo(folder)
        if rec_folder != "":
            folder = rec_folder
        CmdLineDown(settings, folder, args.label).execute()

    @staticmethod
    def init(args: argparse.Namespace):
        remote: str | None = args.remote
        source: str | None = args.source
        folder: str | None = args.folder
        language: str | None = args.language
        database: str | None = args.database
        RepStarter(remote=remote, source=source, folder=folder, language=language, database=database)


    @staticmethod
    def config(args: argparse.Namespace):
        settings = Settings()
        param = ConfigParams()
        param.side = args.side
        param.down = args.down
        param.images = args.images
        param.editor = args.editor
        param.borders = args.borders
        param.timeout = args.timeout
        CmdConfig.execute(settings, param)

    @staticmethod
    def list(_: argparse.Namespace):
        settings = Settings()
        print(str(settings))


class Parser:
    def __init__(self):
        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(prog='tko', description=f'tko version {__version__}')
        self.subparsers = self.parser.add_subparsers(title='subcommands', help='help for subcommand.')

        self.parent_manip = self.create_parent_manip()
        self.parent_basic = self.create_parent_basic()

        self.add_parser_global()
        self.create_parent_basic()
        self.create_parent_manip()
        self.add_parser_tui()
        self.add_parser_run()
        self.add_parser_build()
        self.add_parser_config()
        self.add_parser_rep_tools()
        self.add_parser_rep_actions()
        self.add_parser_diff()


    def add_parser_global(self):
        self.parser.add_argument('-c', metavar='CONFIG_FILE', type=str, help='config json file.')
        self.parser.add_argument('-w', metavar='WIDTH', type=int, help="terminal width.")
        self.parser.add_argument('-v', action='store_true', help='show version.')
        self.parser.add_argument('-m', action='store_true', help='monochromatic.')

    def create_parent_basic(self):
        parent_basic = argparse.ArgumentParser(add_help=False)
        parent_basic.add_argument('--index', '-i', metavar="I", type=int, help='run a specific index.')
        parent_basic.add_argument('--pattern', '-p', metavar="P", type=str, default='@.in @.sol',
                                  help='pattern load/save a folder, default: "@.in @.sol"')
        return parent_basic

    def create_parent_manip(self):
        parent_manip = argparse.ArgumentParser(add_help=False)
        parent_manip.add_argument('--width', '-w', type=int, help="term width.")
        parent_manip.add_argument('--unlabel', '-u', action='store_true', help='remove all labels.')
        parent_manip.add_argument('--number', '-n', action='store_true', help='number labels.')
        parent_manip.add_argument('--sort', '-s', action='store_true', help="sort test cases by input size.")
        parent_manip.add_argument('--pattern', '-p', metavar="@.in @.out", type=str, default='@.in @.sol',
                                  help='pattern load/save a folder, default: "@.in @.sol"')
        return parent_manip

    def add_parser_tui(self):
        parser_r = self.subparsers.add_parser('tui', parents=[self.parent_basic], help='Run using Terminal User Interface.')
        parser_r.add_argument('target_list', metavar='T', type=str, nargs='*', help='solvers, test cases or folders.')
        parser_r.add_argument('--filter', '-f', action='store_true', help='filter solver in temp dir before run')
        parser_r.set_defaults(func=Main.tui)

    def add_parser_run(self):
        parser_r = self.subparsers.add_parser('run', parents=[self.parent_basic], help='Run using raw terminal.')
        parser_r.add_argument('target_list', metavar='T', type=str, nargs='*', help='solvers, test cases or folders.')
        parser_r.add_argument('--filter', '-f', action='store_true', help='filter solver in temp dir before run')

        parser_r.add_argument("--tui", '-t', action='store_true', help='use TUI interface.')
        parser_r.add_argument('--compact', '-c', action='store_true', help='Do not show case descriptions in failures')
        parser_r.add_argument('--eval', '-e', action='store_true', help='Evaluation mode.')

        group_n = parser_r.add_mutually_exclusive_group()
        group_n.add_argument('--quiet', '-q', action='store_true', help='quiet mode, do not show any failure.')
        group_n.add_argument('--all', '-a', action='store_true', help='show all failures.')

        # add an exclusive group for diff mode
        group = parser_r.add_mutually_exclusive_group()
        group.add_argument('--down', '-d', action='store_true', help="diff mode up-to-down.")
        group.add_argument('--side', '-s', action='store_true', help="diff mode side-by-side.")
        parser_r.set_defaults(func=Main.run)

    def add_parser_build(self):
        parser_b = self.subparsers.add_parser('build', parents=[self.parent_manip], help='Build a test target.')
        parser_b.add_argument('target', metavar='T_OUT', type=str, help='target to be build.')
        parser_b.add_argument('target_list', metavar='T', type=str, nargs='+', help='input test targets.')
        parser_b.set_defaults(func=Main.build)

    def add_parser_diff(self):

        parser_b = self.subparsers.add_parser('diff', help='Build a test target.')
        exclusive_group_target = parser_b.add_mutually_exclusive_group(required=True)
        exclusive_group_target.add_argument('--path', '-f', action='store_true', help='targets are paths.')
        exclusive_group_target.add_argument('--text', '-t', action='store_true', help='compare two texts.')
        parser_b.add_argument('target_a', type=str, help='first  target to be compared.')
        parser_b.add_argument('target_b', type=str, help='second target to be compared.')
        exclusive_group = parser_b.add_mutually_exclusive_group()
        exclusive_group.add_argument('--side', '-s', action='store_true', help="diff mode side-by-side.")
        exclusive_group.add_argument('--down', '-d', action='store_true', help="diff mode up-to-down.")
        parser_b.set_defaults(func=Main.diff)

    def add_parser_config(self):
        parser_cfg = self.subparsers.add_parser('config', help='Settings tool.')
        subpar_repo = parser_cfg.add_subparsers(title='subcommands', help='help for subcommand.')

        cfg_reset = subpar_repo.add_parser('reset', help='reset all repositories, folders and values to factory default.')
        cfg_reset.set_defaults(func=CmdRep.reset)

        cfg_set = subpar_repo.add_parser('set', help='set default config values.')

        g_diff = cfg_set.add_mutually_exclusive_group()
        g_diff.add_argument('--side', action='store_true', help='set side_by_side diff mode.')
        g_diff.add_argument('--down', action='store_true', help='set up_to_down   diff mode.')

        cfg_set.add_argument("--editor", metavar="cmd", type=str, help='set editor command.')
        cfg_set.add_argument("--borders", metavar="[0|1]", type=str, help='enable borders.')
        cfg_set.add_argument("--images", metavar="[0|1]", type=str, help='enable images.')
        cfg_set.add_argument("--timeout", metavar="sec", type=int, help='set timeout.')

        cfg_set.set_defaults(func=Main.config)

        cfg_list = subpar_repo.add_parser('list', help='list default config values.')
        cfg_list.set_defaults(func=Main.list)


    def add_parser_rep_tools(self):

        parser_repo = self.subparsers.add_parser('rep', help='Repository validation tools.')

        subpar_repo = parser_repo.add_subparsers(title='subcommands', help='help for subcommand.')
        repo_log = subpar_repo.add_parser("check", help="validates log of the repository.")
        repo_log.add_argument('folder', type=str, help='folder to be checked.')
        repo_log.set_defaults(func=CmdRep.check)

        repo_resume = subpar_repo.add_parser("resume", help="resume log of the repository.")
        repo_resume.add_argument('folder', type=str, help='folder to be checked.')
        repo_resume.set_defaults(func=CmdRep.resume)

        repo_resume = subpar_repo.add_parser("graph", help="resume log of the repository.")
        repo_resume.add_argument('folder', type=str, help='folder to be checked.')
        repo_resume.set_defaults(func=CmdRep.graph)



    def add_parser_rep_actions(self):
        parser_open = self.subparsers.add_parser('open', help='Open a folder with a repository.')
        parser_open.add_argument('folder', type=str, nargs='?', default='.', help='repository folder.')
        parser_open.set_defaults(func=Main.open)

        parser_down = self.subparsers.add_parser('down', help='Down a task from a repository.')
        parser_down.add_argument('folder', type=str, help='repository folder.')
        parser_down.add_argument('label', type=str, help='task label.')
        parser_down.set_defaults(func=Main.down)

        parser_init = self.subparsers.add_parser('init', help='Initialize a repository in a folder.')
        source = parser_init.add_mutually_exclusive_group()
        source.add_argument('--remote', '-r', type=str, help='remote source [fup|ed|poo].')
        source.add_argument('--source', '-s', type=str, help='http url or local file.')
        parser_init.add_argument('--folder', '-f', type=str, help='local directory.')
        parser_init.add_argument('--language', '-l', type=str, help='draft language for the repository.')
        parser_init.add_argument('--database', '-d', type=str, help='define database folder.')
        parser_init.set_defaults(func=Main.init)


def execute(parser: argparse.ArgumentParser, args: argparse.Namespace) -> int:
    settings = Settings()
    if args.w is not None:
        RawTerminal.set_terminal_size(args.w)
    if args.c:
        settings.set_settings_file(args.c)
    settings.load_settings()
    Logger.instance = Logger()

    if args.m:
        AnsiColor.enabled = False
    else:
        AnsiColor.enabled = True
        symbols.set_colors()

    if args.v:
        if args.v:
            print("tko version " + __version__)
            return 0
    else:
        if "func" in args:
            code = args.func(args)
            return code if code is not None else 0
        else:
            parser.print_help()
    return 0

def main():
    try:
        parser = Parser().parser
        args = parser.parse_args()
        execute(parser, args)
        sys.exit(0)
    except KeyboardInterrupt:
        print("\n\nKeyboard Interrupt")
        sys.exit(1)
    except Warning as w:
        print("")
        print(w)
        sys.exit(1)
    # except Exception as e:
    #     print(e)
    #     sys.exit(1)

# NÃO RETIRE ESSA TAG POR CAUSA DO MERGE DO REPLIT
if __name__ == '__main__': # MERGE_INSERT 
    main()



from typing import Optional

from tko.enums.diff_mode import DiffMode
from tko.enums.diff_count import DiffCount

class Param:

    def __init__(self):
        pass

    class Basic:
        def __init__(self):
            self.index: Optional[int] = None
            self.label_pattern: Optional[str] = None
            self.diff_mode = DiffMode.SIDE
            self.diff_count = DiffCount.FIRST
            self.filter: bool = False
            self.compact: bool = False

        def set_index(self, value: Optional[int]):
            self.index= value
            return self

        def set_label_pattern(self, label_pattern: Optional[str]):
            self.label_pattern = label_pattern
            return self
        
        def set_compact(self, value: bool):
            self.compact = value
            return self

        def set_diff_mode(self, value: DiffMode):
            self.diff_mode = value
            return self
    
        def set_filter(self, value: bool):
            self.filter = value
            return self

        def set_diff_count(self, value: DiffCount):
            self.diff_count = value
            return self

    class Manip:
        def __init__(self):
            self.unlabel: bool = False
            self.to_sort: bool = False
            self.to_number: bool = False
        
        def set_unlabel(self, value: bool):
            self.unlabel = value
            return self
        
        def set_to_sort(self, value: bool):
            self.to_sort = value
            return self
        
        def set_to_number(self, value: bool):
            self.to_number = value
            return self



from tko.play.flags import FlagsMan
from tko.play.floating import Floating, FloatingInput, FloatingInputData
from tko.play.floating_manager import FloatingManager
from tko.util.text import Text
from tko.settings.repository import Repository, available_languages
from tko.settings.settings import Settings

class LanguageSetter:
    def __init__(self, settings: Settings, rep: Repository, flagsman: FlagsMan, fman: FloatingManager):
        self.flagsman = flagsman
        self.fman = fman
        self.rep = rep
        self.settings = settings


    def set_language(self):
        options: list[FloatingInputData] = []
        for lang in available_languages:
            options.append(FloatingInputData(TextFunctor(lang), SetLangFunctor(self.settings, self.rep, self.fman, lang)))

        self.fman.add_input(
            FloatingInput(self.settings, "^")
            .set_header(" Escolha a extensão default para os rascunhos ")
            .set_options(options)
            .set_default_index(available_languages.index(self.rep.get_lang()))
            .set_footer(" Escolha e reinicie o tko para aplicar!!!!! ")
        )

class TextFunctor:
    def __init__(self, data: str):
        self.value = data
    
    def __call__(self):
        return Text().add(self.value)

class SetLangFunctor:
    def __init__(self, settings: Settings, rep: Repository, fman: FloatingManager, lang: str):
        self.rep = rep
        self.fman = fman
        self.value = lang
        self.settings = settings

    def __call__(self):
        self.rep.set_lang(self.value.strip())
        self.rep.save_config()
        self.fman.add_input(
            Floating(self.settings, "v>")
            .put_text("")
            .put_text("Linguagem alterada para " + self.value)
            .put_text("")
            .warning()
        )



from tko.game.task import Task

import os
import re

class TaskParser:

    def __init__(self, index_path: str, database_folder: str):
        self.index_path = os.path.abspath(index_path) # path of Repository Root Readme file
        self.database_folder = os.path.abspath(database_folder)
        self.task = Task()

    def __load_xp(self, tags_raw: str):
        tags = [tag.strip() for tag in tags_raw.split(" ")]
        for t in tags:
            if t.startswith("+"):
                self.task.xp = int(t[1:])
                self.task.opt = True
            elif t.startswith("*"):
                self.task.xp = int(t[1:])
                self.task.opt = False
            

    def parse_line(self, line: str, line_num: int = 0) -> Task | None:
        pattern = r'\s*?- \[ \](.*?)\[([^\]]+)\]\(([^)]+)\)(?:\s*<!--(.*?)-->)?'

        match = re.match(pattern, line)
        if match is None:
            return None
        task = self.task
        task.line_number = line_num
        task.line = line
        task.title = match.group(1).strip()
        if task.title != "":
            task.title += " "
        task.title += match.group(2).strip()
        task.title = task.title.replace("`", "")
        self.__load_xp(task.title)

        if match.group(4) is not None:
            self.__load_xp(match.group(4))

        for item in task.title.split(" "):
            if item.startswith("@"):
                task.key = item[1:]
            elif item.startswith("#"):
                task.key = item[1:]
                task.link_type = Task.Types.VISITABLE_URL

        #remove last non alfa char from key
        allowed = "0123456789_abcdefghijklmnopqrstuvwxyz-ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        key = ""
        for c in task.key:
            if c in allowed:
                key += c
            else:
                break
        task.key = key

        link = match.group(3).strip()
        self.select_link_type(link)

        if task.key == "":
            task.key = link
        return task

    def select_link_type(self, link: str):
        task = self.task
        task.link = link

        if task.link_type == Task.Types.VISITABLE_URL:
            # open both url and files
            return

        if self.task.key == "":
            raise Warning(f"Chave não definida para tarefa: {link}")

        if link.startswith("http:") or link.startswith("https:"):
            task.set_folder(os.path.relpath(os.path.join(self.database_folder, task.key)))
            task.link_type = Task.Types.REMOTE_FILE
            return

        if not os.path.isabs(link):
            basedir = os.path.dirname(self.index_path)
            link = os.path.join(basedir, link)
            task.link = os.path.relpath(link)
        # verify if file exists
        # update link using index_path to update de relative path
        if not os.path.isfile(link):
            raise Warning(f"Arquivo não encontrado: {link}")

        # verify if file is inside database_folder/folder
        abs_task_folder = os.path.abspath(os.path.join(self.database_folder, task.key))
        task.set_folder(os.path.relpath(abs_task_folder))
        if link.startswith(abs_task_folder):
            task.link_type = Task.Types.STATIC_FILE
            return

        task.link_type = Task.Types.IMPORT_FILE
        return


from tko.util.text import Text
from tko.play.fmt import Fmt
from tko.play.floating import Floating
from tko.play.input_manager import InputManager
from tko.settings.settings import Settings


class FloatingCalibrate(Floating):
    def __init__(self, settings: Settings):
        super().__init__(settings)
        self.set_header(" Calibrar teclas direcionais ")
        # self.set_text_ljust()
        self.set_footer(" Use Enter para salvar ou ESC para cancelar ")
        self._index = 0
        self._options: list[int] = [settings.app.get_key_left(),
                                   settings.app.get_key_right(),
                                   settings.app.get_key_up(),
                                   settings.app.get_key_down()]
        self._frame.set_border_color("m")
        self._exit_on_action = True
        self.right_dx = 5 # shortcut space
        self.settings = settings
        self.set_key_content()

    def set_key_content(self):
        def format(value: int) -> str:
            if value == 0:
                return "---"
            return str(value)
        self._content = []
        color = "G" if self._index == 0 else ""
        self._content.append(Text().addf(color, "Left  ").addf(color, format(self._options[0])))
        color = "G" if self._index == 1 else ""
        self._content.append(Text().addf(color, "Right ").addf(color, format(self._options[1])))
        color = "G" if self._index == 2 else ""
        self._content.append(Text().addf(color, "Up    ").addf(color, format(self._options[2])))
        color = "G" if self._index == 3 else ""
        self._content.append(Text().addf(color, "Down  ").addf(color, format(self._options[3])))

    # @override
    def get_input(self) -> int:
        self.draw()
        key: int = Fmt.getch()
        key = InputManager.fix_cedilha(Fmt.get_screen(), key)
        
        if key == InputManager.esc:
            self._enable = False
        elif key == ord('\n'):
            self._enable = False
            self.settings.app.set_key_left(self._options[0])
            self.settings.app.set_key_right(self._options[1])
            self.settings.app.set_key_up(self._options[2])
            self.settings.app.set_key_down(self._options[3])
            self.settings.save_settings()
            return -1
        elif key > 128:
            self._options[self._index] = key
            for i in range(len(self._options)):
                if i != self._index and self._options[i] == key:
                    self._options[i] = 0
            self._index += 1
            if self._index >= len(self._options):
                self._index = 0
            self.set_key_content()
        return -1



import os

from tko.util.identifier import Identifier
from tko.util.pattern import PatternLoader
from tko.enums.identifier_type import IdentifierType
from tko.run.unit import Unit
from tko.util.decoder import Decoder

class Writer:

    def __init__(self):
        pass

    @staticmethod
    def to_vpl(unit: Unit):
        text = "case=" + unit.case + "\n"
        text += "input=" + unit.inserted
        text += "output=\"" + unit.get_expected() + "\"\n"
        if unit.grade is None:
            text += "\n"
        else:
            text += "grade reduction=" + str(unit.grade).zfill(3) + "%\n"
        return text

    @staticmethod
    def to_tio(unit: Unit):
        text  = ">>>>>>>> INSERT"
        if unit.case != '':
            text += " " + unit.case
        if unit.grade is not None:
            text += " " + str(unit.grade) + "%"
        text += '\n' + unit.inserted
        text += "======== EXPECT\n"
        text += unit.get_expected()
        if unit.get_expected() != '' and unit.get_expected()[-1] != '\n':
            text += '\n'
        text += "<<<<<<<< FINISH\n\n"
        return text

    @staticmethod
    def save_dir_files(folder: str, pattern_loader: PatternLoader, label: str, unit: Unit) -> None:
        file_source = pattern_loader.make_file_source(label)
        with open(os.path.join(folder, file_source.input_file), "w", encoding="utf-8") as f:
            f.write(unit.inserted)
        with open(os.path.join(folder, file_source.output_file), "w", encoding="utf-8") as f:
            f.write(unit.get_expected())

    @staticmethod
    def save_target(target: str, unit_list: list[Unit], quiet: bool) -> bool:
        def save_dir(_target: str, _unit_list: list[Unit]):
            folder = _target
            pattern_loader = PatternLoader()
            number = 0
            for unit in _unit_list:
                Writer.save_dir_files(folder, pattern_loader, str(number).zfill(2), unit)
                number += 1

        def save_file(_target: str, _unit_list: list[Unit]):
            if _target.endswith(".tio"):
                _new = "\n".join([Writer.to_tio(unit) for unit in _unit_list])
            else:
                _new = "\n".join([Writer.to_vpl(unit) for unit in _unit_list])

            file_exists = os.path.isfile(_target)

            if file_exists:
                _old = Decoder.load(_target)
                if _old == _new:
                    if not quiet:
                        print("no changes in test file")
                    return False

            with open(_target, "w", encoding="utf-8") as f:
                f.write(_new)
                if not quiet:
                    print("file " + _target + " wrote")
                return True

        target_type = Identifier.get_type(target)
        if target_type == IdentifierType.OBI:
            save_dir(target, unit_list)
        elif target_type == IdentifierType.TIO or target_type == IdentifierType.VPL:
            save_file(target, unit_list)
        else:
            print("fail: target " + target + " do not supported for build operation\n")
            return False
        return True



import yaml # type: ignore
from tko.settings.log_action import LogAction
from tko.settings.log_info import LogInfo
import datetime
        
class WeekData:
    def __init__(self, weed_day: str = "", elapsed: int = 0):
        self.weed_day = weed_day
        self.elapsed = elapsed

class WeekListener:
    def __init__(self):
        self.day_actions: dict[str, list[LogInfo]] = {}
        self.log_file: str | None = None
        self.format = '%Y-%m-%d %H:%M:%S'
        self.max_minutes = 60

    def listener(self, action: LogAction, new_entry: bool = False):
        decoder = LogInfo().decode(action)
        day = decoder.timestamp.split(" ")[0]

        if day not in self.day_actions:
            self.day_actions[day] = [decoder]
        else:
            last_action = self.day_actions[day][-1]
            decoder.calc_and_set_elapsed(last_action, self.max_minutes, self.format)
            self.day_actions[day].append(decoder)

    def set_task_file(self, value: str):
        self.log_file = value
        return self

    @staticmethod
    def next_day(day: str) -> str:
        date = datetime.datetime.strptime(day, '%Y-%m-%d')
        return (date + datetime.timedelta(days=1)).strftime('%Y-%m-%d')

    @staticmethod
    def week_day(day: str) -> str:
        date = datetime.datetime.strptime(day, '%Y-%m-%d')
        return date.strftime('%A')

    def resume(self):
        days = sorted(self.day_actions.keys())
        day = days[0]
        last = days[-1]
        while True:
            if day == last:
                break
            if day not in self.day_actions:
                self.day_actions[day] = []
            day = self.next_day(day)

        output: dict[str, WeekData] = {}
        for key in sorted(self.day_actions.keys()):
            entries = self.day_actions[key]
            elapsed = 0 if len(entries) == 0 else entries[-1].get_minutes()
            output[key] = WeekData(weed_day=self.week_day(key), elapsed=elapsed)
        return output
    
    def save_yaml(self):
        if self.log_file is None:
            return
        with open(self.log_file, 'w') as file:
            yaml.dump(self.day_actions, file)



from tko.__main__ import Parser, execute
import pytest
import os

class Compare:
    @staticmethod
    def load_and_save(filename: str, received: str):
        folder = "compare"
        if not os.path.isdir(folder):
            os.mkdir(folder)
        rec = os.path.join(folder, filename + ".rec")
        exp = os.path.join(folder, filename + ".exp")
        open(rec, "w").write(received)
        if os.path.isfile(exp):
            expected = open(exp).read()
        else:
            open(exp, "w").write(received)
            expected = received
        return expected, received
    
    @staticmethod
    def text(capsys: pytest.CaptureFixture[str], file: str, cmd: str):
        Compare.list(capsys, file, cmd.split(" "))

    @staticmethod
    def list(capsys: pytest.CaptureFixture[str], file: str, cmd_list: list[str]):
        parser = Parser().parser
        args = parser.parse_args(cmd_list)
        execute(parser, args)
        expected, received = Compare.load_and_save(file, capsys.readouterr().out) # type: ignore
        assert expected == received



# from typing import override
from tko.util.symbols import symbols
from tko.util.text import Text
from tko.game.tree_item import TreeItem
import enum


class Task(TreeItem):

    class Types(enum.Enum):
        UNDEFINED = 0
        VISITABLE_URL = 1
        STATIC_FILE = 2 # static folder inside database
        REMOTE_FILE = 3 # url link do download file
        IMPORT_FILE = 4 # source folder outside database to import files

        # @override
        def __str__(self):
            return self.name

    # default values for tasks
    approach_max = 6
    autonomy_max = 5
    description_max = 5
    desire_max = 5
    effort_max = 5

    def __init__(self):

        super().__init__()
        self.line_number = 0
        self.line = ""

        self.coverage: int = 0 # valor de 0 a 100
        self.approach: int = 0 # valor de 0 a approach_max
        self.autonomy: int = 0 # valor de 0 a autonomy_max
        self.main_idx: int = 0

        self.howclear: int = 0
        self.howfun: int = 0
        self.howeasy: int = 0

        self.qskills: dict[str, int] = {} # default quest skills
        self.skills: dict[str, int] = {} # local skills
        
        self.xp: int = 0
        self.opt: bool = False
        
        self.link = ""
        self.link_type: Task.Types = Task.Types.UNDEFINED

        self.quest_key = ""
        self.cluster_key = ""

        self.folder: str | None = None # the place where are the test case file
        self.__is_reachable = False
        self.default_min_value = 7 # default min grade to complete task


    def set_reachable(self, reachable: bool):
        self.__is_reachable = reachable
        return self

    def is_reachable(self) -> bool:
        return self.__is_reachable

    def set_folder(self, folder: str):
        self.folder = folder
        return self
    
    def get_folder(self) -> str | None:
        return self.folder
    
    @staticmethod
    def decode_approach_autonomy(value: int) -> tuple[int, int]:
        opts = [(0, 0), (1, 1), (1, 2), (2, 2), (3, 2), (1, 3), (2, 3), (3, 3), (4, 3), (3, 4), (4, 4)]
        autonomy = opts[value][0]
        skill = opts[value][1]
        return autonomy, skill

    def load_from_db(self, value: str):
        if value.startswith("{"):
            value_list = value[1:-1]
            key_values = value_list.split(",")
            for kv in key_values:
                k, val = kv.split(":")
                k = k.strip()
                val = val.strip()
                if k == "cov":
                    self.coverage = int(val)
                elif k == "aut" or k == "appr":
                    self.approach = int(val)
                elif k == "hab" or k == "auto":
                    self.autonomy = int(val)
                elif k == "desc":
                    self.howclear = int(val)
                elif k == "desire":
                    self.howfun = int(val)
                elif k == "effort":
                    self.howeasy = int(val)
                elif k == "idx":
                    self.main_idx = int(val)
        elif ":" not in value:
            self.approach, self.autonomy = Task.decode_approach_autonomy(int(value))
        else:
            v = value.split(":")
            if len(v) == 3:
                self.approach, self.autonomy = Task.decode_approach_autonomy(int(v[0]))
                self.main_idx = int(v[1])
                self.coverage = int(v[2])
            elif len(v) == 4:
                self.coverage = (int(v[0]))
                self.approach = (int(v[1]))
                self.autonomy = (int(v[2]))
                self.main_idx = (int(v[3]))

    def save_to_db(self) -> str:
        return "{" + f"cov:{self.coverage}, appr:{self.approach}, auto:{self.autonomy}, desc:{self.howclear}, desire:{self.howfun}, effort:{self.howeasy}, idx:{self.main_idx}" + "}"
    
    def is_db_empty(self) -> bool:
        return self.approach == 0 and self.autonomy == 0 and self.main_idx == 0 and self.coverage == 0

    def get_prog_color(self, min_value: None | int = None) -> str:
        if min_value is None:
            min_value = self.default_min_value
        prog = self.coverage // 10
        if prog == 0:
            return "c"
        if prog < min_value:
            return "r"
        if prog < 10:
            return "y"
        if prog == 10:
            return "g"
        return "w"  

    def get_prog_symbol(self, min_value: None | int = None) -> Text:
        
        if min_value is None:
            min_value = self.default_min_value
        color = self.get_prog_color(min_value)
        prog = self.coverage // 10
        if prog == 0:
            return Text().add("x")
        if prog < min_value:
            return Text().addf(color, str(prog))
        if prog < 10:
            return Text().addf(color, str(prog))
        if prog == 10:
            return Text().addf(color, symbols.check.text)
        return Text().add("0")

    def get_xp(self) -> int:
        if self.xp == 0:
            return 1
        return self.xp

    def get_percent(self):
        return (self.approach + self.autonomy) * self.coverage // 10
    
    def get_ratio(self) -> float:
        return self.get_percent() / 10.0

    def is_complete(self):
        return self.get_percent() >= 100

    def not_started(self):
        return self.get_percent() == 0
    
    def in_progress(self):
        return self.get_percent() > 0 and self.get_percent() < 100

    def set_coverage(self, coverage: int):
        coverage = int(coverage)
        if coverage >= 0 and coverage <= 100:
            self.coverage = coverage
        else:
            print(f"Progresso inválido: {coverage}")

    def set_approach(self, value: int):
        value = int(value)
        if value >= 0 and value <= Task.approach_max:
            self.approach = value
        else:
            print(f"Autonomia inválida: {value}")

    def set_autonomy(self, value: int):
        value = int(value)
        if value >= 0 and value <= Task.autonomy_max:
            self.autonomy = value
        else:
            print(f"Compreensão inválida: {value}")

    def set_description(self, value: int):
        value = int(value)
        if value >= 0 and value <= Task.description_max:
            self.howclear = value
        else:
            print(f"Descrição inválida: {value}")
    
    def set_desire(self, value: int):
        value = int(value)
        if value >= 0 and value <= Task.desire_max:
            self.howfun = value
        else:
            print(f"Interesse inválido: {value}")

    def set_effort(self, value: int):
        value = int(value)
        if value >= 0 and value <= Task.effort_max:
            self.howeasy = value
        else:
            print(f"Dedicação inválida: {value}")

    # @override
    def __str__(self):
        lnum = str(self.line_number).rjust(3)
        key = "" if self.key == self.title else self.key + " "
        return f"{lnum} grade:{self.approach}:{self.autonomy} key:{key} title:{self.title} skills:{self.skills} remote:{self.link} type:{self.link_type} folder:{self.folder}"

    def has_at_symbol(self):
        return any([s.startswith("@") for s in self.title.split(" ")])



import subprocess
import os
from subprocess import PIPE

class Runner:
    def __init__(self):
        pass

    @staticmethod
    def subprocess_run(cmd: str, input_data: str = "", timeout: None | float = None, folder: str | None = None) -> tuple[int, str, str]:
        try:
            answer = subprocess.run(cmd, cwd=folder, shell=True, input=input_data, stdout=PIPE, stderr=PIPE, text=True, timeout=timeout)
            err = ""
            if answer.returncode != 0:

                err = answer.stderr + Runner.decode_code(answer.returncode)
                # err += "\n" + cmd
            # if running on windows
            if os.name == "nt":
                return answer.returncode, answer.stdout.encode("cp1252").decode("utf-8"), err
            return answer.returncode, answer.stdout, err
        except subprocess.TimeoutExpired:
            err = "fail: processo abortado depois de {} segundos".format(timeout)
            return 1, "", err

    @staticmethod
    def clear_screen():
        if os.name == 'nt':
            os.system('cls')
        else:
            os.system('clear')

    @staticmethod
    def decode_code(return_code: int) -> str:
        code = 128 - return_code
        if code == 127:
            return ""
        if code == 139:
            return "fail: segmentation fault"
        if code == 134:
            return "fail: runtime exception"
        return "fail: execution error code " + str(code)



from tko.game.cluster import Cluster
from tko.game.quest import Quest
from tko.game.task import Task
# from tko.game.graph import Graph

from tko.play.floating_grade import FloatingGrade
from tko.settings.logger import Logger
from tko.settings.settings import Settings

from tko.cmds.cmd_down import CmdDown
from tko.cmds.cmd_run import Run

from tko.util.param import Param

from tko.cmds.cmd_down import DownProblem

from tko.play.fmt import Fmt
from tko.play.floating import Floating
from tko.play.gui import Gui
from tko.play.opener import Opener

from tko.play.tasktree import TaskAction
from typing import Callable
import os
import tempfile
import subprocess


class PlayActions:

    def __init__(self, gui: Gui):
        self.app = Settings().app
        self.settings = Settings()
        self.fman = gui.fman
        self.rep = gui.rep
        self.tree = gui.tree
        self.game = gui.game
        self.graph_opened: bool = False
        self.gui = gui

    def gen_graph_path(self) -> str:
        return os.path.join(self.rep.get_rep_dir(), "graph.png")
        

    def open_link_without_stdout_stderr(self, link: str):
        outfile = tempfile.NamedTemporaryFile(delete=False)
        subprocess.Popen("python3 -m webbrowser -t {}".format(link), stdout=outfile, stderr=outfile, shell=True)

    def get_task_folder(self, task: Task) -> str:
        if task.folder is None:
            raise Exception("Folder não encontrado")
        return task.folder

    def open_code(self):
        obj = self.tree.get_selected_throw()
        if isinstance(obj, Task):
            task: Task = obj
            folder = self.rep.get_task_folder_for_label(task.key)
            if os.path.exists(folder):
                opener = Opener(self.settings).set_fman(self.fman)
                opener.set_target([folder]).set_language(self.rep.get_lang())
                opener.load_folders_and_open()
            else:
                self.fman.add_input(
                    Floating(self.settings, "v>")
                    .put_text("\nO arquivo de código não foi encontrado.\n")
                    .error()
                )
        else:
            self.fman.add_input(
                Floating(self.settings, "v>")
                .put_text("\nVocê só pode abrir o código")
                .put_text("de tarefas baixadas.\n")
                .error()
            )

    def open_link(self):
        obj = self.tree.get_selected_throw()
        if isinstance(obj, Task):
            task: Task = obj
            if task.link_type == Task.Types.VISITABLE_URL or task.link_type == Task.Types.REMOTE_FILE:
                try:
                    self.open_link_without_stdout_stderr(task.link)
                except Exception as _:
                    pass
            self.fman.add_input(
                Floating(self.settings, "v>")
                .set_header(" Abrindo link ")
                .put_text("\n " + task.link + " \n")
                .warning()
            )
        elif isinstance(obj, Quest):
            self.fman.add_input(
                Floating(self.settings, "v>")
                .put_text("\nEssa é uma missão.")
                .put_text("\nVocê só pode abrir o link")
                .put_text("de tarefas.\n")
                .error()
            )
        else:
            self.fman.add_input(
                Floating(self.settings, "v>")
                .put_text("\nEsse é um grupo.")
                .put_text("\nVocê só pode abrir o link")
                .put_text("de tarefas.\n")
                .error()
            )

    # def generate_graph(self):
    #     Graph(self.game).set_path(self.gen_graph_path()).set_opt(False).generate()
    #     path = self.gen_graph_path()
    #     if not self.graph_opened:
    #         opener = Opener(self.settings)
    #         opener.open_files([path])
    #         self.graph_opened = True

    def register_action(self, task: Task):
        Logger.get_instance().record_self_grade(task.key, task.coverage, task.approach, task.autonomy, task.howclear, task.howfun, task.howeasy)

    def evaluate(self):
        obj = self.tree.get_selected_throw()
        
        if isinstance(obj, Task):
            self.fman.add_input(
                FloatingGrade(obj, self.settings, "").set_exit_fn(
                    lambda: self.register_action(obj)
                )
            )
            return

    def down_remote_task(self):

        obj = self.tree.get_selected_throw()
        
        if isinstance(obj, Quest):
            self.fman.add_input(
                Floating(self.settings, "v>")
                .put_text("\nEssa é uma missão.")
                .put_text("\nVocê só pode baixar tarefas.\n")
                .error()
            )
            return
        if isinstance(obj, Cluster):
            self.fman.add_input(
                Floating(self.settings, "v>")
                .put_text("\nEsse é um grupo.")
                .put_text("\nVocê só pode baixar tarefas.\n")
                .error()
            )
        if not isinstance(obj, Task):
            return
        self.__down_remote_task(obj)
    
    def __down_remote_task(self, task: Task) -> None:
        if task.link_type != Task.Types.REMOTE_FILE and task.link_type != Task.Types.IMPORT_FILE:
            self.fman.add_input(
                Floating(self.settings, "v>").put_text("\nEssa não é uma tarefa de baixável.\n").error()
            )
            return
        lang = self.rep.get_lang() 
        down_frame = (
            Floating(self.settings, "v>").warning().set_text_ljust().set_header(" Baixando tarefa ")
        )
        # down_frame.put_text(f"\ntko down {self.rep.alias} {task.key} -l {lang}\n")
        self.fman.add_input(down_frame)

        def fnprint(text: str):
            down_frame.put_text(" " + text)
            down_frame.draw()
            Fmt.refresh()

        cmd_down = CmdDown(rep=self.rep, task_key=task.key, settings=self.settings)
        cmd_down.set_game(self.game)
        cmd_down.set_fnprint(fnprint)
        cmd_down.set_language(lang)
        result = cmd_down.execute()
        if result:
            Logger.get_instance().record_down(task.key)


    def select_task_action(self, task: Task) -> None:
        _, action = self.tree.get_task_action(task)
        if action == TaskAction.BAIXAR:
            self.__down_remote_task(task)
            return
        if action == TaskAction.VISITAR:
            self.open_link()
            return
        folder = task.get_folder()
        if not folder:
            raise Warning("Folder não encontrado")
        self.run_selected_task(task, folder)

    def select_task(self) -> Callable[[], None]:
        obj = self.tree.get_selected_throw()

        if isinstance(obj, Quest) or isinstance(obj, Cluster):
            self.tree.toggle(obj)
            return lambda: None

        if isinstance(obj, Task):
            return lambda: self.select_task_action(obj)
        return lambda: None

        
    def run_selected_task(self, task: Task, task_dir: str) -> None:
        folder = task_dir
        run = Run(settings=self.settings, target_list=[folder], param=Param.Basic())
        run.set_lang(self.rep.get_lang())
        opener = Opener(self.settings).set_language(self.rep.get_lang()).set_target([folder])
        run.set_opener(opener)
        run.set_run_without_ask(False)
        run.set_curses(True)
        run.set_task(self.rep, task)

        run.build_wdir()
        
        if not run.wdir.has_solver():
            DownProblem.create_default_draft(task_dir, self.rep.get_lang())
            msg = Floating(self.settings, "v>").warning()
            msg.put_text("\nNenhum arquivo de código na linguagem {} encontrado.".format(self.rep.get_lang()))
            msg.put_text("\nUm arquivo de rascunho foi criado em {}.".format(task_dir))
            self.fman.add_input(msg)
        run.execute()


import os
from typing import Any
from tko.settings.app_settings import AppSettings
import appdirs  # type: ignore
import yaml #type: ignore

from tko.util.text import Text
from tko.play.colors import Colors
from tko.util.decoder import Decoder
from tko.settings.repository import remove_git_merge_tags

def singleton(class_): # type: ignore
    instances = {}
    def getinstance(*args, **kwargs): # type: ignore
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_] # type: ignore
    return getinstance # type: ignore

@singleton
class Settings:
    CFG_FILE = "settings.yaml"


    def __init__(self):
        self.__remote = "remote"
        self.__appcfg = "appcfg"
        self.__colors = "colors"

        self.dict_alias_remote: dict[str, str] = {}
        self.app = AppSettings()
        self.colors = Colors()

        self.settings_file: str = ""
        self.data: dict[str, Any] = {}

    def set_settings_file(self, path: str):
        self.settings_file = path
        return self

    def get_settings_file(self) -> str:
        if self.settings_file == "":
            self.package_name = "tko"
            default_filename = self.CFG_FILE
            self.settings_file = os.path.join(appdirs.user_data_dir(self.package_name), default_filename) # type: ignore
        
        if not os.path.exists(self.settings_file):
            os.makedirs(os.path.dirname(self.settings_file), exist_ok=True)
        return self.settings_file
    
    def get_settings_backup_file(self) -> str:
        return self.get_settings_file() + ".backup"

    def reset(self):
        self.dict_alias_remote = {}
        self.dict_alias_remote["fup"] = "https://github.com/qxcodefup/arcade/blob/master/Readme.md"
        self.dict_alias_remote["ed"] = "https://github.com/qxcodeed/arcade/blob/master/Readme.md"
        self.dict_alias_remote["poo"] = "https://github.com/qxcodepoo/arcade/blob/master/Readme.md"

        self.app = AppSettings()
        self.colors = Colors()
        return self

    def set_alias_remote(self, alias: str, url_or_path: str):
        if not(url_or_path.startswith("http:") or url_or_path.startswith("https:")):
            url_or_path = os.path.abspath(url_or_path)
        self.dict_alias_remote[alias] = url_or_path
        return self

    def has_alias_remote(self, alias: str) -> bool:
        return alias in self.dict_alias_remote

    def get_alias_remote(self, alias: str) -> str:
        if alias in self.dict_alias_remote:
            return self.dict_alias_remote[alias]
        raise Warning(f"Repositório remoto {alias} não encontrado")
    
    def load_settings(self):
        try:
            settings_file = self.get_settings_file()
            backup_file = self.get_settings_backup_file()

            encoding = Decoder.get_encoding(settings_file)
            with open(settings_file, "r", encoding=encoding) as f:
                lines = f.readlines()
                lines = remove_git_merge_tags(lines)
                data: Any = yaml.safe_load("\n".join(lines))
            if data is None or not isinstance(data, dict):
                with open(backup_file, "r", encoding=encoding) as f:
                    lines = f.readlines()
                    lines = remove_git_merge_tags(lines)
                    data = yaml.safe_load("\n".join(lines))
            if data is None or not isinstance(data, dict):
                raise FileNotFoundError(f"Arquivo de configuração vazio: {settings_file}")
            self.data = data
            self.dict_alias_remote = data.get(self.__remote, {}) # type: ignore
            self.app = AppSettings().from_dict(data.get(self.__appcfg, {})) # type: ignore
            self.colors = Colors().from_dict(data.get(self.__colors, {})) # type: ignore
        except:
            self.reset()
            self.save_settings()
        return self
    
    def save_settings(self):
        file = self.get_settings_file()
        value = {
            self.__remote: self.dict_alias_remote,
            self.__appcfg: self.app.to_dict(),
            self.__colors: self.colors.to_dict()
        }
        backup_file = self.get_settings_backup_file()
        if os.path.exists(file):
            if os.path.exists(backup_file):
                os.remove(backup_file)
            os.rename(file, backup_file)
        with open(file, "w", encoding="utf-8") as f:
            yaml.dump(value, f)
        return self

    # @override
    def __str__(self):
        output: list[str] = []
        output.append(str(Text.format("{g}", "Arquivo de configuração:")))
        output.append("- " + self.get_settings_file())
        output.append("")
        output.append(str(Text.format("{g}", "Repositórios remotos cadastrados:")))
        max_alias = max([len(key) for key in self.dict_alias_remote])
        for key in self.dict_alias_remote:
            output.append("- {} : {}".format(key.ljust(max_alias), self.dict_alias_remote[key]))
        output.append("")
        
        app_str = str(self.app)
        return "\n".join(output) + "\n\n" + app_str



from tko.feno.title import FenoTitle
from tko.feno.jsontools import JsonVPL
from tko.feno.older import Older
from tko.feno.remote_md import Absolute
from tko.feno.html import convert_markdown_to_html
from tko.feno.cases import Cases
from tko.feno.log import Log
from tko.feno.mdpp import Mdpp
from tko.feno.filter import DeepFilter
from tko.util.decoder import Decoder
import subprocess
import argparse
import os
import shutil

def norm_join(*args: str) -> str:
    return os.path.normpath(os.path.join(*args))

class Actions:
    def __init__(self, source_dir: str):
        self.cache = norm_join(source_dir, ".cache")
        self.target = norm_join(self.cache, "mapi.json")
        self.source_dir = source_dir
        self.hook = os.path.basename(os.path.abspath(source_dir))
        self.source_readme = norm_join(self.source_dir, "Readme.md")
        self.remote_readme = norm_join(self.cache, "Readme.md")
        self.target_html = norm_join(self.cache, "q.html")
        self.title = ""
        self.cases = norm_join(self.cache, "q.tio")
        self.mapi_json = norm_join(self.cache, "mapi.json")
        self.cache_src = norm_join(self.cache, "draft")
        self.vpl: JsonVPL | None = None
        self.make_remote: bool = False
        self.use_pandoc: bool = False

    def set_remote(self, make_remote: bool):
        self.make_remote = make_remote
        return self

    def validate(self):
        if self.hook == "node_modules" or self.hook.endswith(".json"):
            return False
        if not os.path.isdir(self.source_dir):
            print(f"\n    fail: {self.source_dir} is not a directory")
            return False
        if not os.path.isfile(self.source_readme):
            print(f"\n    fail: {self.source_readme} not found")
            return False
        return True

    def load_title(self):
        self.title = FenoTitle.extract_title(self.source_readme)

    def create_cache(self):
        if not os.path.exists(self.cache):
            os.makedirs(self.cache)
        return self
    
    def recreate_cache(self):
        shutil.rmtree(self.cache, ignore_errors=True)
        os.makedirs(self.cache)
        return self
    
    def need_rebuild(self):
        if not os.path.exists(self.target):
            return True
        older = Older.find_older([self.source_dir, self.target])
        if older == self.target:
            return False

        Log.resume("Changes ", end="")
        Log.verbose(f"  Changes in {self.source_dir}")
        return True
    
    def remote_md(self):
        Absolute.convert_or_copy_or_print(self.source_readme, self.remote_readme, self.make_remote)
        Log.verbose(f"  RemoteFile: {self.remote_readme}")
    
    def html(self):
        title = FenoTitle.extract_title(self.source_readme)
        convert_markdown_to_html(title, self.remote_readme, self.target_html)
        Log.resume("HTML ", end="")
        Log.verbose(f"  HTML  file: {self.target_html}")

    # uses tko to generate cases file
    def build_cases(self):
        Cases.run(self.cases, self.source_readme, self.source_dir)
        Log.resume("Cases ", end="")
        Log.verbose(f"  Cases file: {self.cases}")

    def copy_drafts(self):
        source_src = norm_join(self.source_dir, "src")
        if os.path.isdir(source_src):
            Log.resume("Drafts ", end="")
            Log.verbose(f"  Drafts dir: {source_src}")
            filter = DeepFilter().set_indent(4)
            filter.copy(source_src, self.cache_src, 5)

    def run_local_sh(self):
        local_sh = norm_join(self.source_dir, "local.sh")
        actual_chdir = os.getcwd()
        if os.path.isfile(local_sh):
            Log.verbose(f"  Execute local.sh")
            os.chdir(self.source_dir)
            subprocess.run("bash local.sh", shell=True)
            os.chdir(actual_chdir)
            Log.resume("Local.sh ", end="")

    def init_vpl(self):
        html_content = Decoder.load(self.target_html)
        # md_content = Decoder.load(self.remote_readme)
        self.vpl = JsonVPL(self.title, html_content)
        self.vpl.set_tests(self.cases)
        if self.vpl.load_drafts(self.cache_src):
            Log.resume("Drafts ", end="")

    def create_mapi(self):
        Decoder.save(self.mapi_json, str(self.vpl) + "\n")
        Log.resume("Mapi ", end="")
        Log.verbose(f"  Mapi  file: {self.mapi_json}")

    def clean(self, erase: bool):
        if erase:
            Log.resume("Cleaning ", end="")
            Log.verbose("  Cleaning  : html and cases files")
            os.remove(self.cases)
            os.remove(self.target_html)
            os.remove(self.remote_readme)

    # run mdpp script on source readme
    def update_markdown(self):
        if Mdpp.update_file(self.source_readme):
            Log.resume("Mdpp ", end="")
            Log.verbose(f"  Mdpp updading")

def build_main(args: argparse.Namespace):
    Log.set_verbose(not args.brief)

    if len(args.targets) == 0:
        print("fail: No targets specified")
        exit(1)

    for target in args.targets:
        hook = os.path.basename(os.path.abspath(target))

        actions = Actions(target)\
                    .set_remote(args.remote)

        if not actions.validate():
            continue

        Log.resume("- " + hook, end=": [ ")
        Log.verbose("- " + hook)

        actions.load_title()
        actions.create_cache()
        actions.update_markdown()

        if not args.check or actions.need_rebuild():
            actions.recreate_cache() # erase .cache
            actions.copy_drafts()
            actions.run_local_sh()
            actions.update_markdown() # se os drafts tiverem mudado o markdown precisa ser atualizado
            actions.remote_md()
            actions.html()
            actions.build_cases()
            actions.init_vpl()
            actions.create_mapi()
            actions.clean(args.erase)

        Log.resume("]")


from tko.game.game import Game
from tko.settings.settings import Settings
from tko.settings.repository import available_languages, Repository
from tko.play.keys import GuiKeys
from tko.play.floating import Floating
from tko.play.floating_calibrate import FloatingCalibrate
from tko.play.fmt import Fmt
from tko.play.input_manager import InputManager
from tko.play.play_config import PlayConfig
from tko.play.floating import Floating
from tko.play.floating_manager import FloatingManager
from tko.play.flags import Flags, FlagsMan
from tko.play.tasktree import TaskTree
from tko.play.gui import Gui
from tko.play.play_actions import PlayActions
from tko.play.flag_functors import FlagFunctor

import curses

class Play:
    def __init__(self, settings: Settings, rep: Repository):
        self.settings = settings
        self.app = settings.app
        self.rep = rep
        self.game: Game = rep.game

        self.exit = False

        self.flagsman = FlagsMan(self.rep.get_flags())
        self.fman = FloatingManager()
        self.tree = TaskTree(self.settings, rep, self.fman)
        self.gui = Gui(tree=self.tree, flagsman=self.flagsman, fman=self.fman)

        # if len(self.rep.get_tasks()) == 0:
        #     self.gui.show_help()

        self.first_loop = True
        self.graph_ext = ""

        self.actions = PlayActions(self.gui)
        self.play_palette = PlayConfig(self.actions)

    def set_need_update(self):
        self.gui.set_need_update()

    def save_to_json(self):
        self.tree.save_on_rep()
        self.rep.set_flags(self.flagsman.get_data())
        self.settings.save_settings()
        self.rep.save_config()

    def send_quit_msg(self):
        def set_exit():
            self.exit = True

        self.fman.add_input(
            Floating(self.settings).put_text("\nAté a próxima\n").set_exit_fn(set_exit).warning()
        )

    def toggle_skills(self):
        Flags.skills.toggle()

    def make_callback(self) -> InputManager:
        cman = InputManager()

        cman.add_str(GuiKeys.key_quit, self.send_quit_msg)
        cman.add_int(InputManager.esc, self.send_quit_msg)
        # cman.add_int(curses.KEY_BACKSPACE, self.send_quit_msg)

        cman.add_str(GuiKeys.up, self.tree.move_up)
        cman.add_str(GuiKeys.up2, self.tree.move_up)
        key_up = self.app.get_key_up()
        if key_up != 0:
            cman.add_int(key_up, self.tree.move_up)

        cman.add_str(GuiKeys.down, self.tree.move_down)
        cman.add_str(GuiKeys.down2, self.tree.move_down)
        key_down = self.app.get_key_down()
        if key_down != 0:
            cman.add_int(key_down, self.tree.move_down)

        cman.add_str(GuiKeys.left, self.tree.arrow_left)
        cman.add_str(GuiKeys.left2, self.tree.arrow_left)
        key_left = self.app.get_key_left()
        if key_left != 0:
            cman.add_int(key_left, self.tree.arrow_left)

        cman.add_str(GuiKeys.right, self.tree.arrow_right)
        cman.add_str(GuiKeys.right2, self.tree.arrow_right)
        key_right = self.app.get_key_right()
        if key_right != 0:
            cman.add_int(key_right, self.tree.arrow_right)

        cman.add_str(GuiKeys.calibrate, lambda: self.fman.add_input(FloatingCalibrate(self.settings)))
        
        cman.add_str(GuiKeys.activate, lambda: self.actions.select_task()) # type: ignore
        cman.add_str(GuiKeys.open_url, self.actions.open_link)
        cman.add_str(GuiKeys.down_task, self.actions.down_remote_task)
        cman.add_str(GuiKeys.edit, lambda: self.actions.open_code())
        cman.add_str(GuiKeys.expand, self.tree.process_expand_all)
        cman.add_str(GuiKeys.collapse, self.tree.process_collapse_all)
        cman.add_str(GuiKeys.borders, self.app.toggle_borders)
        cman.add_str(GuiKeys.images, self.app.toggle_images)
        cman.add_str(GuiKeys.hidden, self.app.toggle_hidden)
        cman.add_str(GuiKeys.set_lang_drafts, self.gui.language.set_language)
        cman.add_int(curses.KEY_BACKSPACE, self.actions.evaluate)
        cman.add_str(GuiKeys.grade_play, self.actions.evaluate)
        if Flags.admin:
            cman.add_str(GuiKeys.mass, self.tree.mass_mark)
    
        cman.add_str(GuiKeys.key_help, self.gui.show_help)
        
        for flag in self.flagsman.flags.values():
            cman.add_str(flag.get_keycode(), FlagFunctor(flag))

        cman.add_str(GuiKeys.search, self.gui.search.toggle_search)
        cman.add_str(GuiKeys.palette, self.play_palette.command_pallete)

        return cman
        
    def send_char_not_found(self, key: int):
        exclude_str = [ord(v) for v in [" ", "a", "d", "\n", GuiKeys.up, GuiKeys.down, GuiKeys.left, GuiKeys.right]]
        exclude_int = [ -1, InputManager.esc] + InputManager.backspace_list
        if key in exclude_int + exclude_str:
            return
        self.fman.add_input( Floating(self.settings, "v>").error().put_text(f"Tecla char:{chr(key)}, code:{key}, não reconhecida") )


    def main(self, scr: curses.window):
        InputManager.fix_esc_delay()
        curses.curs_set(0)  # Esconde o cursor
        Fmt.init_colors()  # Inicializa as cores
        Fmt.set_scr(scr)  # Define o scr como global

        while True:
            self.tree.update_tree(admin_mode=Flags.admin.is_true() or self.gui.search.search_mode)
            self.fman.draw_warnings()
            cman = self.make_callback()
            self.gui.show_items()

            if self.fman.has_floating():
                value: int = self.fman.get_input()
            else:
                value = scr.getch()
                value = InputManager.fix_cedilha(scr, value)

            if self.exit:
                break

            if self.gui.search.search_mode:
                self.gui.search.process_search(value)
            elif cman.has_int_key(value):
                callback = cman.exec_call(value)
                if callback is not None:
                    return callback
            else:
                self.send_char_not_found(value)

            self.tree.reload_sentences()
            self.save_to_json()
            if self.first_loop:
                self.first_loop = False

    def check_lang_in_text_mode(self):
        lang = self.rep.get_lang()
        if lang == "":
            options = available_languages
            print("\nLinguagem padrão ainda não foi definida.\n")
            while True:
                print("Escolha entre as opções a seguir ", end="")
                print("[" + ", ".join(options) + "]", ":", end=" ")
                lang = input()
                if lang in options:
                    break
            self.rep.set_lang(lang)


    def play(self):
        self.check_lang_in_text_mode()

        while True:
            output = curses.wrapper(self.main)
            if output is None:
                return
            else:
                output()



import enum
import json
import os
from tko.util.decoder import Decoder

def json_norm_join(*args: str):
    return os.path.normpath(os.path.join(*args))

class FileType(enum.Enum):
    SHOW = "show" # visível para o aluno como required
    HIDE = "hide" # arquivos extra utilizados na compilação, mas escondidos dos alunos
    EXEC = "exec" # arquivos a serem copiados para a pasta que o código será executado

# Format used to send additional files to VPL
class JsonFile:
    def __init__(self, name: str, contents: str, ftype: FileType):
        self.name: str = name
        self.contents: str = contents
        self.ftype: FileType = ftype

    def to_json(self) -> dict[str, str]:
        return {
            "name" : self.name,
            "contents": self.contents,
            "type": self.ftype.value,
        }

    # @override
    def __str__(self):
        return self.name + ":" + self.contents + ":" + str(self.ftype.value)

class JsonVPL:
    hide_extension = ".hide"
    exec_extension = ".exec"

    def __init__(self, title: str, description: str):
        self.title: str = title
        self.description: str = description
        self.upload: list[JsonFile] = []
        self.draft: dict[str, list[JsonFile]] = {}
    
    def set_tests(self, exec_file: str):
        data = Decoder.load(exec_file)
        self.upload.append(JsonFile("vpl_evaluate.cases", data, FileType.HIDE))
        return self
    
    def add_draft(self, extension: str, exec_file: str, ftype: FileType):
        if extension not in self.draft:
            self.draft[extension] = []
        filename = exec_file.split(os.sep)[-1]
        data = Decoder.load(exec_file)
        jfile = JsonFile(filename, data, ftype)
        self.draft[extension].append(jfile)
        return self

    def to_json(self) -> str:
        json_dict = {
            "title": self.title,
            "description": self.description,
            "upload": [x.to_json() for x in self.upload],
            "draft": {k: [x.to_json() for x in v] for k, v in self.draft.items()},
        }
        return json.dumps(json_dict, indent=4, ensure_ascii=True)
    
    def load_drafts(self, cache_draft: str):
        found = False
        if not os.path.isdir(cache_draft):
            return False
        folders: list[str] = os.listdir(cache_draft)
        folders = [f for f in folders if os.path.isdir(os.path.join(cache_draft, f))]
        for lang in folders:
            lang_path = json_norm_join(cache_draft, lang)
            hide_files: list[str] = []
            hide_files_config = os.path.join(lang_path, JsonVPL.hide_extension)
            if os.path.isfile(hide_files_config):
                with open(hide_files_config) as f:
                    hide_files = [x.strip() for x in f.read().splitlines()]
            exec_files: list[str] = []
            exec_files_config = os.path.join(lang_path, JsonVPL.exec_extension)
            if os.path.isfile(exec_files_config):
                with open(exec_files_config) as f:
                    exec_files = [x.strip() for x in f.read().splitlines()]

            for file in os.listdir(lang_path):
                if file.endswith(JsonVPL.hide_extension) or file.endswith(JsonVPL.exec_extension):
                    continue
                ftype = FileType.SHOW
                if file in hide_files:
                    ftype = FileType.HIDE
                elif file in exec_files:
                    ftype = FileType.EXEC
                file_path = json_norm_join(lang_path, file)
                self.add_draft(lang, file_path, ftype)
                found = True
        return found

    # @override
    def __str__(self):
        return self.to_json()



from tko.game.task import Task
from tko.play.floating import Floating
from tko.play.fmt import Fmt
from tko.play.input_manager import InputManager
from tko.util.symbols import symbols
from tko.util.text import Text
from tko.play.keys import GuiKeys
from tko.settings.settings import Settings
# from typing import override

import curses


class FloatingGrade(Floating):
    def __init__(self, task: Task, settings: Settings, _align: str =""):
        super().__init__(settings, _align)
        self.settings = settings
        self._task = task
        self._line = 0
        self.set_text_ljust()
        self._frame.set_border_color("g")
        self.set_header_text(Text.format("{y/}", " Utilize os direcionais e teclas - e + para  marcar "))
        self.set_footer_text(Text.format("{y/}", " Pressione Enter para confirmar "))

        self.grades_index = [task.coverage // 10, task.approach, task.autonomy, task.howclear, task.howfun, task.howeasy]
        self.coverage = ["x", "1", "2", "3", "4", "5", "6", "7", "8", "9", "✓"]
        self.coverage_msg = [
            Text().addf("m", "Não ").addf("y", "fiz"),
            Text().addf("m", "Fiz ").addf("y", "10%"),
            Text().addf("m", "Fiz ").addf("y", "20%"),
            Text().addf("m", "Fiz ").addf("y", "30%"),
            Text().addf("m", "Fiz ").addf("y", "40%"),
            Text().addf("m", "Fiz ").addf("y", "50%"),
            Text().addf("m", "Fiz ").addf("y", "60%"),
            Text().addf("m", "Fiz ").addf("y", "70%"),
            Text().addf("m", "Fiz ").addf("y", "80%"),
            Text().addf("m", "Fiz ").addf("y", "90%"),
            Text().addf("m", "Fiz ").addf("y", "100%")]
        self.approach = [x.text for x in [symbols.approach_x, symbols.approach_e, symbols.approach_d, symbols.approach_c, symbols.approach_b, symbols.approach_a, symbols.approach_s]]
        self.approach_msg = [
            Text().addf("m", "Não fiz"),
            Text().addf("m", "Peguei o código ").addf("y", "pronto").addf("m", " e estudei ele"),
            Text().addf("m", "Fiz com ajuda de ").addf("y", "IA").addf("m", "(copilot|gpt|outros)"),
            Text().addf("m", "Fiz").addf("y"," guiado ").addf("m", "por aula|vídeo|colega|monitor"),
            Text().addf("y", "(Re)Fiz").addf("m", " os códigos").addf("m", " pesquisando"),
            Text().addf("y", "(Re)Fiz sozinho").addf("m", " sem consultar algoritmos"),
            Text().addf("y", "Fiz de primeira").addf("m", " sem consultar algoritmos")
            ]
        self.autonomy = [x.text for x in [symbols.autonomy_x, symbols.autonomy_e, symbols.autonomy_d, symbols.autonomy_c, symbols.autonomy_b, symbols.autonomy_a]]
        self.autonomy_msg = [
            Text().addf("m", "Não fiz        ").addf("y", "                         "),
            Text().addf("m", "Consigo refazer ").addf("y", ">= 30% sem consulta      "),
            Text().addf("m", "Consigo refazer ").addf("y", ">= 50% sem consulta      "),
            Text().addf("m", "Consigo refazer ").addf("y", ">= 70% sem consulta      "),
            Text().addf("m", "Consigo refazer ").addf("y", ">= 90% sem consulta      "),
            Text().addf("m", "Consigo refazer ").addf("y", "  100% sem consulta      ")
            ]

        self.description = [x.text for x in [symbols.desire_x, 
                                             symbols.desire_e, 
                                             symbols.desire_d, 
                                             symbols.desire_c, 
                                             symbols.desire_b, 
                                             symbols.desire_a]]
 
        self.description_msg = [
            Text().addf("m", "Sem opinião               "),
            Text().addf("m", "A descrição da atividades ").addf("y", "é confusa     "),
            Text().addf("m", "A descrição da atividades ").addf("y", "é insuficiente"),
            Text().addf("m", "A descrição da atividades ").addf("y", "é razoável    "),
            Text().addf("m", "A descrição da atividades ").addf("y", "é adequada    "),
            Text().addf("m", "A descrição da atividades ").addf("y", "é excelente   ")
            ]
        
       
        self.desire = [x for x in self.description]
        self.desire_msg = [
            Text().addf("m", "Sem opinião        ").addf("y", "                     "),
            Text().addf("m", "A atividade parece ").addf("y", "muito desinteressante"),
            Text().addf("m", "A atividade parece ").addf("y", "desinteressante      "),
            Text().addf("m", "A atividade parece ").addf("y", "indiferente          "),
            Text().addf("m", "A atividade parece ").addf("y", "interessante         "),
            Text().addf("m", "A atividade parece ").addf("y", "muito interessante   "),
            ]

        self.effort = [x for x in self.description]
        self.effort_msg = [
            Text().addf("m", "Sem opinião          ").addf("y", "                   "),
            Text().addf("m", "Resolver a atividade ").addf("y", "foi muito difícil  "),
            Text().addf("m", "Resolver a atividade ").addf("y", "foi difícil        "),
            Text().addf("m", "Resolver a atividade ").addf("y", "foi razoável       "),
            Text().addf("m", "Resolver a atividade ").addf("y", "foi fácil          "),
            Text().addf("m", "Resolver a atividade ").addf("y", "foi muito fácil    "),
            ]

        self.grades_value = [self.coverage, self.approach, self.autonomy, self.description, self.desire, self.effort]
        self.grades_msg = [self.coverage_msg, self.approach_msg, self.autonomy_msg, self.description_msg, self.desire_msg, self.effort_msg]


    def update_content(self):
        self._content = []
        self._content.append(Text())
        self._content.append(Text().add(" ======= Pontue de acordo com a última fez que você (re)fez a tarefa do zero ======="))
        self._content.append(Text().add(" ---------------------------------- Obrigatório ------------------------------------"))
        coverage_question = "Cobertura: Quanto entregou?"
        approach_question = "Abordagem: Como foi feito ?"
        autonomy_question = "Autonomia: Quanto domina  ?"
        description_question = "Descrição: É bem escrita  ?"
        desire_question   = "Interesse: É interessante ?"
        effort_question   = "Dedicação: É fácil        ?"

        coverage_text = Text().add(" ").addf("Y" if self._line == 0 else "", coverage_question).add("  ")
        for i, c in enumerate(self.coverage):
            coverage_text.addf("C" if i == self.grades_index[0] else "", c)
        coverage_text.add("  ").add(self.coverage_msg[self.grades_index[0]])
        self._content.append(coverage_text)
        # self._content.append(Text())
        approach_text = Text().add(" ").addf("Y" if self._line == 1 else "", approach_question).add("  ")
        for i, c in enumerate(self.approach):
            approach_text.addf("G" if i == self.grades_index[1] else "", c).add("")
            if c == "x":
                approach_text.add("  ")
            if c == "A":
                approach_text.add("  ")
            if c == "S":
                approach_text.add(" ")
        
        approach_text.add(" ").add(self.approach_msg[self.grades_index[1]])
        self._content.append(approach_text)
        # self._content.append(Text())
        autonomy_text = Text().add(" ").addf("Y" if self._line == 2 else "", autonomy_question).add("  ")
        for i, c in enumerate(self.autonomy):
            autonomy_text.addf("yW" if i == self.grades_index[2] else "", c).add(" ")
        autonomy_text.add(" ").add(self.autonomy_msg[self.grades_index[2]])
        self._content.append(autonomy_text)

        self._content.append(Text().add(" ------------------------------------ Opcional -------------------------------------"))

        description_text = Text().add(" ").addf("Y" if self._line == 3 else "", description_question).add("  ")
        for i, c in enumerate(self.description):
            description_text.addf("yW" if i == self.grades_index[3] else "", c).add(" ")
        description_text.add(" ").add(self.description_msg[self.grades_index[3]])
        self._content.append(description_text)

        desire_text = Text().add(" ").addf("Y" if self._line == 4 else "", desire_question).add("  ")
        for i, c in enumerate(self.desire):
            desire_text.addf("yW" if i == self.grades_index[4] else "", c).add(" ")
        desire_text.add(" ").add(self.desire_msg[self.grades_index[4]])
        self._content.append(desire_text)

        effort_text = Text().add(" ").addf("Y" if self._line == 5 else "", effort_question).add("  ")
        for i, c in enumerate(self.effort):
            effort_text.addf("yW" if i == self.grades_index[5] else "", c).add(" ")
        effort_text.add(" ").add(self.effort_msg[self.grades_index[5]])
        self._content.append(effort_text)

        self._content.append(Text())

    # @override
    def draw(self):
        self._set_default_header()
        self._set_default_footer()
        self.setup_frame()
        self._frame.draw()
        self.update_content()
        self.write_content()

    def change_task(self):
        self._task.set_coverage(self.grades_index[0] * 10)
        self._task.set_approach(self.grades_index[1])
        self._task.set_autonomy(self.grades_index[2])
        self._task.set_description(self.grades_index[3])
        self._task.set_desire(self.grades_index[4])
        self._task.set_effort(self.grades_index[5])

    # @override
    def get_input(self) -> int:
        self.draw()
        key: int = Fmt.getch()
        key = InputManager.fix_cedilha(Fmt.get_screen(), key)
        if key == self.settings.app.get_key_up() or key == ord(GuiKeys.up) or key == ord(GuiKeys.up2):
            self._line = max(self._line - 1, 0)
        elif key == self.settings.app.get_key_down() or key == ord(GuiKeys.down) or key == ord(GuiKeys.down2):
            self._line = min(self._line + 1, len(self.grades_index) - 1)
        elif key == self.settings.app.get_key_left() or key == ord(GuiKeys.left) or key == ord(GuiKeys.left2):
            self.grades_index[self._line] = max(self.grades_index[self._line] - 1, 0)
            self.change_task()
        elif key == self.settings.app.get_key_right() or key == ord(GuiKeys.right) or key == ord(GuiKeys.right2):
            self.grades_index[self._line] = min(self.grades_index[self._line] + 1, len(self.grades_value[self._line]) - 1)
            self.change_task()
        elif key == ord("0"):
            for i in range(len(self.grades_index)):
                self.grades_index[i] = len(self.grades_value[i]) - 1
        elif key == ord("-"):
            for i in range(len(self.grades_index)):
                self.grades_index[i] = max(self.grades_index[i] - 1, 0)
        elif key == ord("+") or key == ord("="):
            for i in range(len(self.grades_index)):
                self.grades_index[i] = min(self.grades_index[i] + 1, len(self.grades_value[i]) - 1)
        elif key == ord('\n') or key == curses.KEY_BACKSPACE or key == InputManager.esc:
            self._enable = False
            if self._exit_fn is not None:
                self._exit_fn()
        return -1


from tko.util.text import Text

class TreeItem:
    def __init__(self):
        self.sentence: Text = Text()
        self.key: str = ""
        self.title: str = ""

    def get_key(self):
        return self.key

    def get_title(self):
        return self.title
    
    def get_sentence(self):
        return self.sentence



from typing import Callable, Any
import os
import urllib.request
import urllib.error
import json
import shutil
import tempfile


from tko.game.task import Task
from tko.settings.repository import Repository
from tko.settings.settings import Settings
from tko.game.game import Game
from tko.util.remote_url import RemoteUrl
from tko.cmds.cmd_build import CmdBuild
from tko.util.param import Param
from tko.util.decoder import Decoder
from tko.down.drafts import Drafts
from tko.settings.repository import available_languages

class CmdLineDown:
    def __init__(self, settings: Settings, folder: str, task_key: str):
        self.settings = settings
        self.folder = folder
        self.task_key = task_key
        self.rep = Repository(folder)
        
    def execute(self):
        if not self.rep.has_local_config_file():
            print("O parâmetro para o comando tko down deve a pasta onde você iniciou o repositório.")
            print("Navegue ou passe o caminho até a pasta do repositório e tente novamente.")
            return False
        self.rep.load_config().load_game()
        CmdDown(self.rep, self.task_key, self.settings).execute()
        return True


class CmdDown:
    def __init__(self, rep: Repository, task_key: str, settings: Settings):
        self.rep = rep
        self.task_key = task_key
        self.settings = settings
        self.game: Game | None = None
        self.language: str | None = None
        self.destiny_folder: str = ""
        self.readme_path: str = ""
        self.cache_url: str = ""
        self.mapi_file: str = ""

    def set_fnprint(self, fnprint: Callable[[str], None]):
        DownProblem.fnprint = fnprint
        return self

    def set_game(self, game: Game):
        self.game = game
        return self
    
    def set_language(self, language: str):
        self.language = language
        return self

    def remove_empty_destiny_folder(self):
        if len(os.listdir(self.destiny_folder)) == 0:
                os.rmdir(self.destiny_folder)

    def download_readme(self, readme_remote_url: RemoteUrl) -> bool:
        try:
            DownProblem.down_readme(self.readme_path, readme_remote_url)
            return True
        except urllib.error.HTTPError:
            DownProblem.fnprint("falha: Arquivo de descrição não encontrado")
            return False
        except urllib.error.URLError:
            DownProblem.fnprint("falha: Não consegui baixar a atividade, verifique sua internet")
            return False

    def download_mapi(self) -> bool:
        try:
            mapi_url = self.cache_url + "mapi.json"
            urllib.request.urlretrieve(mapi_url, self.mapi_file)
            return True
        except urllib.error.HTTPError:
            return False
        except urllib.error.URLError:
            DownProblem.fnprint("falha: não consegui baixar a atividade, verifique sua internet")
            return False

    def build_cases_from_readme(self, destiny_folder: str):
        cases_tio_target = os.path.join(destiny_folder, "cases.tio")
        param = Param.Manip()
        cb = CmdBuild(cases_tio_target, [self.readme_path], param)
        cb.set_quiet(True)
        cb.execute()

    def download_from_url(self, task_source: str) -> bool:
        readme_remote_url = RemoteUrl(task_source)
        remote_url = readme_remote_url.get_raw_url()
        self.cache_url = os.path.dirname(remote_url) + "/.cache/"
        self.destiny_folder = self.rep.get_task_folder_for_label(self.task_key)
        self.readme_path =  os.path.join(self.destiny_folder, "Readme.md")
        self.mapi_file = os.path.join(self.destiny_folder, "mapi.json")

        backup_folder = DownProblem.backup_and_create_problem_folder(self.destiny_folder)
        if not self.download_readme(readme_remote_url):
            self.remove_empty_destiny_folder()
            return False
        
        lang = self.check_and_select_language()
        
        if self.download_mapi():
            encoding = Decoder.get_encoding(self.mapi_file)
            with open(self.mapi_file, encoding=encoding) as f:
                loaded_json = json.load(f)
            os.remove(self.mapi_file)
            DownProblem.unpack_problem_files(loaded_json, self.destiny_folder)
            if not DownProblem.unpack_json_drafts(loaded_json, self.destiny_folder, lang):
                DownProblem.create_default_draft(self.destiny_folder, lang)
        else:
            self.build_cases_from_readme(self.destiny_folder)
            DownProblem.create_default_draft(self.destiny_folder, lang)
        DownProblem.fnprint("")
        DownProblem.fnprint("Atividade baixada com sucesso")
        if backup_folder != "":
            if CmdDown.folder_equals(self.destiny_folder, backup_folder):
                DownProblem.fnprint("Pastas {} e {} são iguais".format(os.path.basename(self.destiny_folder), os.path.basename(backup_folder)))
                DownProblem.fnprint("Removendo pasta de backup {}".format(os.path.basename(backup_folder)))
                shutil.rmtree(backup_folder, ignore_errors=True)

        return True

    @staticmethod
    def folder_equals(folder1: str, folder2: str) -> bool:
        """ Compare two folders and return if they have the same files with the same content """
        if not os.path.exists(folder1) or not os.path.exists(folder2):
            return False
        if len(os.listdir(folder1)) != len(os.listdir(folder2)):
            return False
        for file in os.listdir(folder1):
            path1 = os.path.join(folder1, file)
            path2 = os.path.join(folder2, file)
            if not os.path.isfile(path1) or not os.path.isfile(path2):
                return False
            with open(path1, "rb") as f1, open(path2, "rb") as f2:
                if f1.read() != f2.read():
                    return False
        return True


    def download_from_external_file(self, task_source: str) -> bool:
        if not os.path.exists(task_source):
            DownProblem.fnprint("Arquivo fonte não encontrado")
        self.cache_url = os.path.join(os.path.dirname(task_source), ".cache")
        self.destiny_folder = self.rep.get_task_folder_for_label(self.task_key)
        self.readme_path =  os.path.join(self.destiny_folder, "Readme.md")
        self.mapi_file = os.path.join(self.cache_url, "mapi.json")
        backup_folder = DownProblem.backup_and_create_problem_folder(self.destiny_folder)
        # Decoder.save(self.readme_path, Decoder.load(task_source))
        DownProblem.compare_and_save(Decoder.load(task_source), self.readme_path)

        lang = self.check_and_select_language()
        
        if os.path.isfile(self.mapi_file):
            encoding = Decoder.get_encoding(self.mapi_file)
            with open(self.mapi_file, encoding=encoding) as f:
                loaded_json = json.load(f)
            DownProblem.unpack_problem_files(loaded_json, self.destiny_folder)
            if not DownProblem.unpack_json_drafts(loaded_json, self.destiny_folder, lang):
                if not backup_folder:
                    DownProblem.create_default_draft(self.destiny_folder, lang)
        else:
            self.build_cases_from_readme(self.destiny_folder)
            DownProblem.create_default_draft(self.destiny_folder, lang)
        DownProblem.fnprint("")
        DownProblem.fnprint("Atividade carregada com sucesso")
        return True

    def execute(self) -> bool:
        task = self.rep.game.get_task(self.task_key)
        if task.link_type == Task.Types.REMOTE_FILE:
            return self.download_from_url(task.link)
        if task.link_type == Task.Types.IMPORT_FILE:
            return self.download_from_external_file(task.link)
        DownProblem.fnprint("falha: link para atividade não possui link para download")
        return False

    def check_and_select_language(self) -> str:
        language_def = self.rep.get_lang()

        if self.language is None:
            if language_def != "":
                self.language = language_def
            else:
                print("Escolha uma extensão para os rascunhos: [{}]: ".format(", ".join(available_languages)), end="")
                self.language = input()
        return self.language

class DownProblem:
    fnprint: Callable[[str], None] = print

    @staticmethod
    def __create_file(content: str, path: str, label: str=""):
        Decoder.save(path, content)
        DownProblem.fnprint(DownProblem.folder_and_file(path) + " " + label)

    @staticmethod
    def unpack_json_drafts(loaded: dict[str, Any], destiny: str, lang: str) -> bool:
        found = False
        if "draft" in loaded:
            if lang in loaded["draft"]:
                for file in loaded["draft"][lang]:
                    path = os.path.join(destiny, file["name"])
                    DownProblem.__create_file(file["contents"], path, "(Rascunho)")
                    found = True

        return found

    @staticmethod
    def unpack_problem_files(loaded: dict[str, Any], destiny: str):
        if "upload" in loaded:
            for file in loaded["upload"]:
                name = file["name"]
                if name == "vpl_evaluate.cases":
                    name = "cases.tio"
                path = os.path.join(destiny, name)
                DownProblem.compare_and_save(file["contents"], path)
        if "keep" in loaded:
            for file in loaded["keep"]:
                name = file["name"]
                if name == "vpl_evaluate.cases":
                    name = "cases.tio"
                path = os.path.join(destiny, name)
                DownProblem.compare_and_save(file["contents"], path)

    @staticmethod
    def folder_and_file(path: str) -> str:
        """ Return the folder and file name of the path """
        folder = os.path.basename(os.path.dirname(path))
        file = os.path.basename(path)
        return os.path.join(folder, file)

    @staticmethod
    def  compare_and_save(content: str, path: str):
        if not os.path.exists(path):
            Decoder.save(path, content)
            DownProblem.fnprint(DownProblem.folder_and_file(path) + " (Novo)")
        else:
            path_content = Decoder.load(path)
            if path_content != content:
                DownProblem.fnprint(DownProblem.folder_and_file(path) + " (Atualizado)")
                Decoder.save(path, content)
            else:
                DownProblem.fnprint(DownProblem.folder_and_file(path) + " (Inalterado)")

    @staticmethod
    def down_readme(readme_path: str,  remote_url: RemoteUrl):
        temp_file = tempfile.mktemp()
        remote_url.download_absolute_to(temp_file)
        content = Decoder.load(temp_file)
        DownProblem.compare_and_save(content, readme_path)
    
    @staticmethod
    def backup_and_create_problem_folder(destiny: str) -> str:
        """ If destiny exists, rename it to destiny+1, destiny+2, etc.
        If backup was created, return the new name.
        If not, return empty string.
        """
        new_destiny = ""
        if os.path.exists(destiny): # move folder to 
            count = 1
            while True:
                new_destiny = "{}+{}".format(destiny, count)
                if not os.path.exists(new_destiny):
                    shutil.move(destiny, new_destiny)
                    DownProblem.fnprint("Pasta {} já existe".format(os.path.basename(destiny)))
                    DownProblem.fnprint("Renomeando para {}".format(os.path.basename(new_destiny)))
                    break
                else:
                    count += 1
        nova = "" if new_destiny == "" else " nova"
        DownProblem.fnprint("Criando{} pasta {}".format(nova, os.path.basename(destiny)))
        os.makedirs(destiny, exist_ok=True)
        DownProblem.fnprint("")
        return new_destiny

    @staticmethod
    def create_default_draft(destiny: str, language: str):
        filename = "draft."
        draft_path = os.path.join(destiny, filename + language)

        if not os.path.exists(draft_path):
            with open(draft_path, "w", encoding="utf-8") as f:
                if language in Drafts.drafts:
                    f.write(Drafts.drafts[language])
                else:
                    f.write("")
            DownProblem.fnprint(DownProblem.folder_and_file(draft_path) + " (Vazio)")
        else:
            DownProblem.fnprint(DownProblem.folder_and_file(draft_path) + " (Não sobrescrito)")



from tko.util.param import Param
from tko.run.wdir import Wdir
from tko.run.writer import Writer

class CmdBuild:

    def __init__(self, target_out: str, source_list: list[str], param: Param.Manip):
        self.target_out = target_out
        self.source_list = source_list
        self.param = param
        self.quiet = False


    def set_quiet(self, value: bool):
        self.quiet = value
        return self

    def execute(self):
        try:
            wdir = Wdir().set_sources(self.source_list).build()
            wdir.manipulate(self.param)
            Writer.save_target(self.target_out, wdir.get_unit_list(), quiet=self.quiet)
        except FileNotFoundError as e:
            print(str(e))
            return False
        return True



from __future__ import annotations
from tko.settings.history_file import HistoryFile
from tko.settings.log_action import LogAction
from tko.settings.listener_daily import DailyListener
from tko.settings.listener_task import TaskListener
from tko.settings.listener_week import WeekListener
from tko.settings.listener_cache import CacheListener
from tko.settings.repository import Repository

class Logger:
    instance: None | Logger = None

    @staticmethod
    def get_instance() -> Logger:
        if Logger.instance is None:
            Logger.instance = Logger()
        return Logger.instance

    def __init__(self, repository: Repository | None = None):
        self.rep: Repository | None = repository
        self.last_hash: str | None = None
        self.history: None | HistoryFile = None
        self.daily = DailyListener()
        self.tasks = TaskListener()
        self.week = WeekListener()
        self.cache = CacheListener()

    def set_log_files(self, log_file: str, track_folder: str) -> Logger:
        if self.history is not None:
            return self
        self.tasks.set_track_folder(track_folder)
        self.history = HistoryFile(log_file, [self.daily.listener, 
                                              self.tasks.listener, 
                                              self.week.listener, 
                                              self.cache.listener])
        # resume = self.week.resume()
        # print("\n".join([f"{x.weed_day} {x.elapsed}" for x in resume.values()]))
        return self
    
    def get_history_file(self) -> HistoryFile:
        if self.history is None:
            raise Warning("fail: History File not set")
        return self.history

    def check_log_file_integrity(self) -> list[str]:
        entries = self.get_history_file().get_entries()
        if len(entries) == 0:
            return []
        hash = entries[0].hash
        output: list[str] = []

        for i in range(1, len(entries)):
            calculated_hash = LogAction.generate_hash(entries[i], hash)
            if calculated_hash != entries[i].hash:
                output.append(f"Hash mismatch line {i + 1}: {str(entries[i])}")
            hash = calculated_hash
        return output

    def record_pick(self, task_key: str):
        self.__record_event(LogAction.Type.PICK, task_key)

    def record_back(self, task_key: str):
        self.__record_event(LogAction.Type.BACK, task_key)

    def record_down(self, task_key: str):
        self.__record_event(LogAction.Type.DOWN, task_key)

    def record_compilation_execution_error(self, task_key: str):
        self.__record_event(LogAction.Type.FAIL, task_key)
    
    def record_test_result(self, task_key: str, result: int):
        self.__record_event(LogAction.Type.TEST, task_key, str(result))
    
    def record_file_alteration(self, task_key: str, line_count: int):
        self.__record_event(LogAction.Type.SIZE, task_key, str(line_count))

    def record_freerun(self, task_key: str):
        self.__record_event(LogAction.Type.FREE, task_key)

    def record_self_grade(self, task_key: str, coverage: int, approach: int, autonomy: int, clear: int, fun: int, easy: int):
        self.__record_event(LogAction.Type.SELF, task_key, "{" + f"c:{coverage},a:{approach},s:{autonomy},clear:{clear},fun:{fun},easy:{easy}" + "}")

    def record_open(self):
        self.__record_event(LogAction.Type.OPEN)
    
    def record_quit(self):
        self.__record_event(LogAction.Type.QUIT)

    def __record_event(self, action: LogAction.Type, task_key: str = "", payload: str = ""):
        if self.history is None:
            return
        self.get_history_file().append_new_action(action, task_key, payload)



from __future__ import annotations
from tko.settings.log_action import LogAction
from tko.game.task import Task
import datetime

class LogInfo:
    def __init__(self):
        self.timestamp = ""
        self.type: str = LogAction.Type.NONE.value
        self.key = ""
        self.coverage: int = -1
        self.approach: int = -1
        self.autonomy: int = -1
        self.howclear: int = -1
        self.howfun: int = -1
        self.howeasy: int = -1
        self.elapsed: datetime.timedelta = datetime.timedelta(0)
        self.attempts: int = 0
        self.lines: int = -1

    def get_minutes(self) -> int:
        return int(self.elapsed.total_seconds() / 60)

    def set_timestamp(self, value: str):
        self.timestamp = value
        return self

    def set_type(self, value: str):
        self.type = value
        return self
    
    def set_key(self, value: str):
        self.key = value
        return self
    
    def set_attempts(self, value: int):
        self.attempts = value
        return self
    
    def set_coverage(self, value: int):
        self.coverage = value
        return self
    
    def set_autonomy(self, value: int):
        self.approach = value
        return self
    
    def set_skill(self, value: int):
        self.autonomy = value
        return self
    
    def set_elapsed(self, value: datetime.timedelta):
        self.elapsed = value
        return self

    def set_lines(self, value: int):
        self.lines = value
        return self
    
    def calc_and_set_elapsed(self, last: LogInfo, limit_minutes: int, format: str):
        last_time = datetime.datetime.strptime(last.timestamp, format)
        current_time = datetime.datetime.strptime(self.timestamp, format)
        diff = current_time - last_time
        if diff.total_seconds() < limit_minutes * 60:
            elapsed = last.elapsed + diff
        else:
            elapsed = last.elapsed
        self.elapsed = elapsed
        return self
        
    def to_dict(self) -> dict[str, str]:
        return {
            "type": self.type,
            "key": self.key,
            "coverage": str(self.coverage),
            "approach": str(self.approach),
            "autonomy": str(self.autonomy),
            "howclear": str(self.howclear),
            "howfun": str(self.howfun),
            "howeasy": str(self.howeasy),
            "elapsed": str(int(self.elapsed.total_seconds() / 60)),
            "attempts": str(self.attempts),
            "lines": str(self.lines)
        }

    def __str__(self) -> str:
        output = f'time:{self.timestamp}, type:{self.type}, key:{self.key}'
        if self.coverage != -1:
            output += f', c:{self.coverage}'
        if self.approach != -1:
            output += f', a:{self.approach}'
        if self.autonomy != -1:
            output += f', s:{self.autonomy}'
        if self.howclear != -1:
            output += f', clear:{self.howclear}'
        if self.howfun != -1:
            output += f', fun:{self.howfun}'
        if self.howeasy != -1:
            output += f', easy:{self.howeasy}'
        if self.elapsed.total_seconds() > 0:
            minutes = int(self.elapsed.total_seconds() / 60)
            output += f', e:{minutes}'
        if self.attempts > 0:
            output += f', att:{self.attempts}'
        if self.lines != -1:
            output += f', lines:{self.lines}'
        return output
    
    def decode(self, action: LogAction):
        self.timestamp = action.timestamp
        self.type = action.type_value
        self.key = action.task_key
        if action.type_value == LogAction.Type.TEST.value:
            self.load_from_test(action.payload)
        elif action.type_value == LogAction.Type.PROG.value:
            self.load_from_prog(action.payload)
        elif action.type_value == LogAction.Type.SELF.value:
            self.load_from_self(action.payload)
        elif action.type_value == LogAction.Type.SIZE.value:
            self.load_from_size(action.payload)
        return self

    def load_from_test(self, payload: str):
        try:
            self.coverage = int(payload)
        except:
            pass
    
    def load_from_prog(self, payload: str):
        self.coverage = int(payload)
        return
    
    def load_from_size(self, payload: str):
        self.lines = int(payload)
        return
    
    def load_from_self(self, payload: str):
        if len(payload) == 1:
            self.approach, self.autonomy = Task.decode_approach_autonomy(int(payload))
            return
    
        if len(payload) == 2:
            self.approach = int(payload[0])
            self.autonomy = int(payload[1])
            return
        
        if len(payload) == 3 and payload[0] == "0":
            self.approach = int(payload[1])
            self.autonomy = int(payload[2])
            return
        
        if payload[0] == "{":
            payload = payload[1:-1]
            values = payload.split(",")
            kv: dict[str, str] = {}
            for svalue in values:
                k, v = svalue.split(":")
                kv[k.strip()] = v.strip()
            self.coverage = int(kv.get("c", -1))
            self.approach = int(kv.get("a", -1))
            self.autonomy = int(kv.get("s", -1))
            self.howclear = int(kv.get("clear", -1))
            self.howfun = int(kv.get("fun", -1))
            self.howeasy = int(kv.get("easy", -1))
            return
        
        raise Exception(f"Invalid SELF payload: {payload}")
            
    


from tko.run.unit import Unit
from tko.enums.execution_result import ExecutionResult
from tko.util.runner import Runner
from tko.run.solver_builder import SolverBuilder

class UnitRunner:

    def __init__(self):
        pass

    # run a unit using a solver and return if the result is correct
    @staticmethod
    def run_unit(solver: SolverBuilder, unit: Unit, timeout: None | float = None) -> ExecutionResult:
        exec, _ = solver.get_executable()
        if solver.has_compile_error():
            unit.set_received(exec.get_error_msg().get_str())
            return ExecutionResult.COMPILATION_ERROR
        cmd, folder = exec.get_command()
        if folder == "":
            folder = None
        if timeout == 0:
            timeout = None
        return_code, stdout, stderr = Runner.subprocess_run(cmd, unit.inserted, timeout, folder)
        if return_code != 0:
            unit.set_received(stdout + stderr)
            return ExecutionResult.EXECUTION_ERROR
        unit.set_received(stdout)
        if unit.get_received() == unit.get_expected():
            return ExecutionResult.SUCCESS
        return ExecutionResult.WRONG_OUTPUT


from __future__ import annotations

class TaskBasic:
    def __init__(self, key: str = ""):
        self.key = key
        self.coverage = 0
        self.autonomy = 0
        self.skill = 0
        self.timestamp: str = ""

    def set_coverage(self, value: int):
        if value >= 0:
            self.coverage = value
        return self

    def set_approach(self, value: int):
        if value >= 0:
            self.autonomy = value
        return self
    
    def set_autonomy(self, value: int):
        if value >= 0:
            self.skill = value
        return self

    def set_timestamp(self, timestamp: str):
        self.timestamp = timestamp
        return self

    # @override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TaskBasic):
            return False
        return self.coverage == other.coverage and self.autonomy == other.autonomy and self.skill == other.skill

    # @override
    def __str__(self):
        return "{" + f'k:{self.key}, c:{self.coverage}, a:{self.autonomy}, s:{self.skill}' + "}"


from urllib.request import urlopen
from tko import __version__


class CheckVersion:
    link = "https://raw.githubusercontent.com/senapk/tko/master/src/tko/__init__.py"
    
    buffer: str | None = None
    init: bool = False

    def __init__(self):

        self.version: str = __version__
        if not CheckVersion.init:
            CheckVersion.buffer = self.get_latest_version()
            CheckVersion.init = True
        self.latest_version: str | None = CheckVersion.buffer

    def is_updated(self):
        if self.latest_version is None:
            return True
        if self.version != self.latest_version:
            major, minor, patch = [int(x) for x in self.version.split(".")]
            latest_major, latest_minor, latest_patch = [int(x) for x in self.latest_version.split(".")]
            if major < latest_major or (major == latest_major and minor < latest_minor) or (major == latest_major and minor == latest_minor and patch < latest_patch):
                # print(f"Sua versão do  tko ({self.version}) está desatualizada.")
                # print(f"A última versão é a {self.latest_version}.")
                # print(Text.format("Atualize com o comando {y}.", "pipx upgrade tko"))
                return False
        return True

    def get_latest_version(self):
        try:
            with urlopen(self.link) as f:
                for line in f:
                    if b"__version__" in line:
                        return line.decode().split('"')[1]
        except:
            return None


# ignore missing import 
from typing import Any
import chardet
import os

class Decoder:

    @staticmethod
    def load(file_path: str) -> str:
        if not os.path.exists(file_path):

            raise Exception("Arquivo {} não encontrado".format(os.path.abspath(file_path)))
        with open(file_path, "rb") as file:
            raw_data = file.read()
            enc_dict: dict[str, Any] = chardet.detect(raw_data) # type: ignore
            encoding: str = enc_dict["encoding"] or "utf-8" # type: ignore
            if enc_dict["confidence"] < 0.90:
                encoding = "utf-8"
            try:
                content = raw_data.decode(encoding)  # type: ignore
            except UnicodeDecodeError as _:
                content = raw_data.decode(enc_dict["encoding"] or "utf-8") # type: ignore
            return content.replace('\r\n', '\n')

    @staticmethod
    def save(file_path: str, content: str):
        base_path = os.path.dirname(file_path)
        if base_path != "" and not os.path.exists(base_path):
            os.makedirs(base_path)
        with open(file_path, "w", encoding="utf-8", newline='\n') as file:
            file.write(content)

    @staticmethod
    def get_encoding(file_path: str) -> str:
        with open(file_path, "rb") as file:
            raw_data = file.read()
            enc_dict = chardet.detect(raw_data)
            encoding = enc_dict["encoding"]
            if enc_dict["confidence"] < 0.90:
                encoding = "utf-8"
            return encoding or "utf-8"